import re
import shutil
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any
from collections.abc import Callable, Iterable

import pytest

from testlib.fixtures.env import ManagedEnv
from testlib.fixtures.formatter import BalancedTemplater


class Fileish(ABC):
    """
    Baseclass, which allows files to be copied declaratively
    """

    @abstractmethod
    def copy_to(self, path: Path, origin: Path) -> None:
        """
        Copies this file to the given TempDir
        :param path: TempDir for the test
        :param origin: Directory the tests originates in. Used to adjust relative paths
        """


class _ByContentFileish(Fileish, ABC):
    def __init__(self, mode: int | None = None):
        self.mode = mode

    @abstractmethod
    def get_content(self, origin: Path) -> str:
        """
        Returns the content, which should be present in the current path
        :return: content as a string
        """

    def copy_to(self, path: Path, origin: Path) -> None:
        path.write_text(self.get_content(origin))
        if self.mode is not None:
            path.chmod(self.mode)


class File(_ByContentFileish):
    def __init__(self, file_contents: str, mode: int | None = None):
        """
        Declares a file by its content
        :param file_contents: content of the file as a string
        :param mode: Optionally change the mode of the file (e.g. to executable)
        """
        super().__init__(mode)
        self.file_contents = file_contents

    def get_content(self, _: Path) -> str:
        return self.file_contents


class CopyFile(Fileish):
    def __init__(self, source: str | Path):
        """
        Declares a file as a copy of an existing file
        :param source: Path to the file to be copied
        """
        self.source = source

    def copy_to(self, path: Path, origin: Path):
        orig = self.source if isinstance(self.source, Path) else origin / self.source
        shutil.copyfile(orig, path)


class CopyTree(Fileish):
    def __init__(
        self, tree_base: str | Path, ignore: Callable[[str, list[str]], Iterable[str]] | None = None
    ):
        """
        Declares a folder as a copy of an existing folder
        :param tree_base: base folder of the tree being copied
        """
        self.tree_base = tree_base
        self.ignore = ignore

    def copy_to(self, path: Path, origin: Path):
        orig = self.tree_base if isinstance(self.tree_base, Path) else origin / self.tree_base
        shutil.copytree(orig, path, dirs_exist_ok=True, ignore=self.ignore)


class CopyTemplate(_ByContentFileish):
    def __init__(self, template: str | Path, values: dict[str, Any], mode: int | None = None):
        """
        Declares a file as an initiated version of the given file template
        :param template: source template's file name. Parameters formatted as `{key_name}` are replaced by corresponding values
        :param values: dictionary of key_name and value to be replaced in the template.
        :param mode: Optionally change the mode of the file (e.g. to executable)
        """
        self.template = template
        self.values = values
        self.content: str | None = None

        super().__init__(mode)

    def get_content(self, origin: Path) -> str:
        template_path = self.template if isinstance(self.template, Path) else origin / self.template
        template_content = template_path.read_text()

        self.content = BalancedTemplater(template_content).substitute(**self.values)

        return self.content


class EnvTemplate(_ByContentFileish):
    def __init__(self, template_content: str, mode: int | None = None):
        """
        Creates a File with the given content,
        but replaces all @ENV_VARIABLE_NAME@ with the according value
        :param template_content: String of the file content with placeholders
        """
        self.template_content = template_content
        self.env = None
        super().__init__(mode)

    def set_env(self, env: ManagedEnv):
        self.env = env

    def get_content(self, _: Path) -> str:
        keys = re.findall(r"@(\w+)@", self.template_content)
        return BalancedTemplater(self.template_content).substitute({k: self.env[k] for k in keys})


class Symlink(Fileish):
    def __init__(self, target: str):
        """
        Declares a file as a symlink to a different location.

        NOTE: due to limitations of symlinks on Windows, tests using this might be flaky and fail!!
        :param target: Path to the target where the symlink should be pointing
        """
        self.target = target

    def copy_to(self, path: Path, _origin: Path):
        path.symlink_to(self.target)


class AssetSymlink(Fileish):
    def __init__(self, source: str):
        """
        Declares a file as a symlink to a local asset file.

        NOTE: due to limitations of symlinks on Windows, tests using this might be flakey and fail!!
        :param source: Path to the source where the symlink should be pointing.
            Paths are relative to the current test's module.
        :raise ValueError: When the given source path is absolute
        """
        self.source = source

    def copy_to(self, path: Path, origin: Path):
        if Path(self.source).is_absolute():
            msg = "absolute paths are not allowed"
            raise ValueError(msg)
        target_path = origin / self.source
        path.symlink_to(target_path)


type FileDeclaration = dict[str, Fileish | "FileDeclaration"]


def merge_file_declaration(a: FileDeclaration, b: FileDeclaration) -> FileDeclaration:
    result = {}
    for key in a.keys() | b.keys():
        if (key in a) ^ (key in b):
            result[key] = a.get(key) or b.get(key)
            continue
        if isinstance(a[key], Fileish) or isinstance(b[key], Fileish):
            msg = "Cannot merge files; got two different values for the same path"
            raise ValueError(msg, key)
        try:
            result[key] = merge_file_declaration(a[key], b[key])
        except ValueError as e:
            msg, path = e.args
            raise ValueError(msg, f"{key}/{path}")

    return result


def _init_files(
    files: FileDeclaration, current_folder: Path, caller_path: Path, env: ManagedEnv
) -> None:
    """
    This internal function is needed because one cannot call a fixture directly since pytest 4.0
    """

    for name, definition in files.items():
        destination = current_folder / name
        if isinstance(definition, Fileish):
            if isinstance(definition, EnvTemplate):
                definition.set_env(env)
            definition.copy_to(destination, caller_path)
        else:
            # Initialize subdirectory
            destination.mkdir()
            _init_files(definition, destination, caller_path, env)


def with_files(*files: FileDeclaration) -> Callable[[Any], Callable[[Any], None]]:
    def decorator(func: Callable[[Any], None]) -> Callable[[Any], None]:
        return pytest.mark.usefixtures("files")(
            pytest.mark.parametrize("files", files, indirect=True)(func)
        )

    return decorator


@pytest.fixture
def files(env: ManagedEnv, request: pytest.FixtureRequest) -> Path:
    """
    Initializes the given files into the TempDir of the test.
    This ensures all necessary files and only those are present
    To use this add `@with_files(list_of_your_files_to_test, more_files_to_test)` above your test.
    The test is run once for each of the sets of files provided as the second argument.
    Each Set of files should be of the :py:type:`FileDeclaration` type

    :param env: environment to get the HOME path from
    :param request: Fixture information provided by pytest, used to parametrize the files
    :return: Path to where the files were created
    """
    home = env.dirs.home
    if hasattr(request, "param"):
        _init_files(request.param, home, request.path.parent, env)
    return home



from testlib.fixtures.nix import Nix
from testlib.fixtures.file_helper import with_files, CopyTree, AssetSymlink
from testlib.fixtures.snapshot import Snapshot
from pathlib import Path
import os


@with_files(
    {
        "flake": CopyTree("assets/metadata/flake"),
        "metadata.out": AssetSymlink("assets/metadata/metadata.out"),
    }
)
def test_metadata(nix: Nix, files: Path, snapshot: Snapshot):
    nix.settings.add_xp_feature("nix-command", "flakes")
    nix.env.dirs.nix_store_dir = "/nix/store"
    nix.env["TZ"] = "UTC"
    nix.env["LC_ALL"] = "C.UTF-8"

    os.utime(files / "flake/flake.nix", (1000, 1000))
    os.utime(files / "flake/flake.lock", (1000, 1000))
    os.utime(files / "flake", (1000, 1000))

    # We use NIX_STORE_DIR which causes unstable paths. This is goofy. We can
    # just use `--store` which sets `rootDir` and does not have this problem.
    result = (
        nix.nix(["flake", "metadata", "--store", nix.env.dirs.real_store_dir, files / "flake"])
        .run()
        .ok()
    )
    assert snapshot("metadata.out") == result.stdout_s.replace(str(files), "$root")



from typing import Callable
import urllib.parse
import re
import functools
import subprocess
import dataclasses

S3_HOST = 's3-admin.afnix.fr'
S3_USER = 'lix-releng'

DEFAULT_STORE_URI_BITS = {
    'region': 'global',
    'endpoint': 's3.afnix.fr',
    'want-mass-query': 'true',
    'write-nar-listing': 'true',
    'ls-compression': 'zstd',
    'narinfo-compression': 'zstd',
    'compression': 'zstd',
    'parallel-compression': 'true',
}


@dataclasses.dataclass
class DockerTarget:
    registry_path: str
    """Registry path without the tag, e.g. ghcr.io/lix-project/lix"""

    tags: list[str]
    """List of tags this image should take. There must be at least one."""

    @staticmethod
    def resolve(item: str, version: str, major: str) -> str:
        """
        Applies templates:
        - version: the Lix version e.g. 2.90.0
        - major: the major Lix version e.g. 2.90
        """
        return item.format(version=version, major=major)

    def registry_name(self) -> str:
        [a, _, _] = self.registry_path.partition('/')
        return a


@dataclasses.dataclass
class RelengEnvironment:
    name: str
    colour: Callable[[str], str]

    cache_store_overlay: dict[str, str]
    cache_bucket: str
    releases_bucket: str
    docs_bucket: str
    # We don't want to call functions that require computation (such as guess_gerrit_remote()) before they are needed
    git_repo: Callable[[], str]
    git_repo_is_gerrit: bool
    s3_endpoint: str
    s3_region: str
    s3_ssh_host: str | None
    s3_ssh_user: str | None

    docker_targets: list[DockerTarget]

    def cache_store_uri(self):
        qs = DEFAULT_STORE_URI_BITS.copy()
        qs.update(self.cache_store_overlay)
        return self.cache_bucket + "?" + urllib.parse.urlencode(qs)


SGR = '\x1b['
RED = '31;1m'
GREEN = '32;1m'
BLUE = '34;1m'
RESET = '0m'


def sgr(colour: str, text: str) -> str:
    return f'{SGR}{colour}{text}{SGR}{RESET}'


LOCAL = RelengEnvironment(
    name='local',
    colour=functools.partial(sgr, BLUE),
    docs_bucket='s3://local-docs',
    cache_bucket='s3://local-cache',
    cache_store_overlay={'secret-key': 'local.key', 'endpoint': 'localhost:3900', 'scheme': 'http'},
    releases_bucket='s3://local-releases',
    git_repo=lambda: '../releng-target',
    git_repo_is_gerrit=False,
    docker_targets=[],
    s3_endpoint = 'http://localhost:3900',
    s3_region = 'garage',
    s3_ssh_host = None,
    s3_ssh_user = None,
)


STAGING = RelengEnvironment(
    name='staging',
    colour=functools.partial(sgr, GREEN),
    docs_bucket='s3://docs.staging.lix.systems',
    cache_bucket='s3://cache.staging.lix.systems',
    cache_store_overlay={'secret-key': 'staging.key'},
    releases_bucket='s3://releases.staging.lix.systems',
    git_repo=lambda: 'ssh://git@git.lix.systems/lix-project/lix-releng-staging',
    git_repo_is_gerrit=False,
    docker_targets=[
        # latest will be auto tagged if appropriate
        DockerTarget('git.lix.systems/lix-project/lix-releng-staging',
                     tags=['{version}', '{major}']),
        DockerTarget('ghcr.io/lix-project/lix-releng-staging',
                     tags=['{version}', '{major}']),
    ],
    s3_endpoint = 'https://s3.afnix.fr',
    s3_region = 'garage',
    s3_ssh_host = S3_HOST,
    s3_ssh_user = S3_USER,
)

GERRIT_REMOTE_RE = re.compile(r'^ssh://(\w+@)?gerrit.lix.systems:2022/lix$')


def guess_gerrit_remote():
    """
    Deals with people having unknown gerrit username.
    """
    out = [
        x.split()[1] for x in subprocess.check_output(
            ['git', 'remote', '-v']).decode().splitlines()
    ]
    return next(x for x in out if GERRIT_REMOTE_RE.match(x))


PROD = RelengEnvironment(
    name='production',
    colour=functools.partial(sgr, RED),
    docs_bucket='s3://docs.lix.systems',
    cache_bucket='s3://cache.lix.systems',
    # FIXME: we should decrypt this with age into a tempdir in the future, but
    # the issue is how to deal with the recipients file. For now, we should
    # just delete it after doing a release.
    cache_store_overlay={'secret-key': 'prod.key'},
    releases_bucket='s3://releases.lix.systems',
    git_repo=guess_gerrit_remote,
    git_repo_is_gerrit=True,
    docker_targets=[
        # latest will be auto tagged if appropriate
        DockerTarget('git.lix.systems/lix-project/lix',
                     tags=['{version}', '{major}']),
        DockerTarget('ghcr.io/lix-project/lix', tags=['{version}', '{major}']),
    ],
    s3_endpoint = 'https://s3.afnix.fr',
    s3_region = 'global',
    s3_ssh_host = S3_HOST,
    s3_ssh_user = S3_USER,
)

ENVIRONMENTS = {
    'local': LOCAL,
    'staging': STAGING,
    'production': PROD,
}


@dataclasses.dataclass
class S3Credentials:
    name: str
    id: str
    secret_key: str



from testlib.fixtures.nix import Nix
import aiohttp.web as web
from testlib.fixtures.http_server import http_server
import pytest
import json


@pytest.mark.parametrize("registry", [None, "vendored"])
def test_default_registry(nix: Nix, registry: str | None):
    nix.settings.flake_registry = registry
    items = nix.nix(["registry", "list"], flake=True).run().ok().stdout_s.splitlines()
    # Make sure the vendored registry contains the correct amount.
    assert len(items) == 37
    # sanity check, contains the important ones
    assert any(item.startswith("global flake:nixpkgs") for item in items)
    assert any(item.startswith("global flake:home-manager") for item in items)


online_registry = {
    "flakes": [
        {
            "from": {"type": "indirect", "id": "nixpkgs"},
            "to": {"type": "github", "owner": "NixOS", "repo": "nixpkgs"},
        },
        {
            "from": {"type": "indirect", "id": "private-flake"},
            "to": {"type": "github", "owner": "fancy-enterprise", "repo": "private-flake"},
        },
    ],
    "version": 2,
}


class TestOnlineRegistry:
    async def get_registry(self, _req: web.Request) -> web.Response:
        return web.Response(text=json.dumps(online_registry))

    def test_online_registry(self, nix: Nix):
        app = web.Application()
        app.add_routes([web.get("/flake-registry.json", self.get_registry)])

        with http_server(app) as httpd:
            nix.settings.flake_registry = f"http://localhost:{httpd.port}/flake-registry.json"

            result = nix.nix(["registry", "list"], flake=True).run().ok()
            assert result.stdout_s.splitlines() == [
                "global flake:nixpkgs github:NixOS/nixpkgs",
                "global flake:private-flake github:fancy-enterprise/private-flake",
            ]

            # make sure there's a warning
            assert (
                "config option flake-registry referring to a URL is deprecated and will be removed"
                in result.stderr_s
            )



from pathlib import Path
from collections.abc import Callable

import pytest

from testlib.fixtures.nix import Nix
from testlib.fixtures.snapshot import Snapshot
from testlib.utils import functional2_base_folder


@pytest.mark.parametrize(
    ("expr", "exit_code", "exit_code_depr"),
    [
        ("00012.3", 1, 0),
        ("0a", 1, 0),
        ("0https://a", 1, 0),
        ("0.0.0", 1, 0),
        ("""foo"1"2""", 1, 0),
        ("0x10", 1, 0),
        ("0.", 1, 1),
        ("1.", 0, 0),
        ("0.a", 1, 0),
        ("1.a", 1, 0),
        (""" 0."" """, 1, 0),
        (""" 1."" """, 1, 0),
        # test against false positives
        ("(0)(0)", 0, 0),
        ('a("")', 0, 0),
        ("(a).a", 0, 0),
        # not deprecated atm but probably should be in the future (FIXME piegames; 2026-01-30)
        ("(a).0", 0, 0),
        # unrelated syntax errors which don't trigger the deprecation
        ("00.", 1, 1),
    ],
)
def test_whitespace(
    nix: Nix,
    expr: str,
    exit_code: int,
    exit_code_depr: int,
    snapshot: Callable[[str], Snapshot],
    files: Path,
):
    expr = expr.strip()
    for i, expr in enumerate((f"({expr})", f"[{expr}]")):
        f_expr = expr.replace("/", "-")
        for f in [
            f"{f_expr}.out.exp",
            f"{f_expr}.err.exp",
            f"{f_expr}-depr.out.exp",
            f"{f_expr}-depr.err.exp",
        ]:
            (files / f).symlink_to(functional2_base_folder / "lang" / "parser-token-whitespace" / f)

        full_expr = f"with {{}}; {expr}"
        res = (
            nix.nix_instantiate(
                ["--parse", "--extra-deprecated-features", "url-literals", "-E", full_expr]
            )
            .run()
            .expect(exit_code)
        )
        assert snapshot(f"{f_expr}.out.exp") == res.stdout_s
        assert snapshot(f"{f_expr}.err.exp") == res.stderr_s

        res = (
            nix.nix_instantiate(
                [
                    "--parse",
                    "--extra-deprecated-features",
                    "tokens-no-whitespace url-literals",
                    "-E",
                    full_expr,
                ]
            )
            .run()
            .expect(exit_code_depr)
        )
        assert snapshot(f"{f_expr}-depr.out.exp") == res.stdout_s
        assert snapshot(f"{f_expr}-depr.err.exp") == res.stderr_s



from testlib.fixtures.nix import Nix
from testlib.fixtures.git import Git
from testlib.fixtures.file_helper import with_files, File
from testlib.utils import get_global_asset_pack
from pathlib import Path


@with_files({"flake1": get_global_asset_pack(".git"), "input": File("input")})
def test_absolute_path(nix: Nix, files: Path, git: Git):
    nix.settings.add_xp_feature("nix-command", "flakes")

    flake = files / "flake1"
    (flake / "flake.nix").write_text(f"""{{
        outputs = {{ self }}: {{ x = builtins.readFile {files}/input; }};
    }}""")

    git(flake, "add", "flake.nix")
    git(flake, "commit", "-m", "Initial")

    assert nix.nix(["eval", "--impure", "--json", f"{flake}#x"]).run().ok().stdout_s == '"input"\n'



"""
Safe (read-only) environment access. This exists entirely so that we can ban
os.environ in our codebase.
"""

import collections.abc
import os
from typing import TypeVar, overload
from collections.abc import Iterator

_T = TypeVar("_T")


class ReadonlyDict(collections.abc.Mapping[str, str]):
    def __init__(self, inner: os._Environ[str]):
        self.inner = inner

    def __getitem__(self, key: str) -> str:
        return self.inner[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self.inner)

    def __len__(self) -> int:
        return len(self.inner)

    @overload
    def get(self, key: str, /) -> str | None: ...
    @overload
    def get(self, key: str, /, default: str | _T) -> str | _T: ...

    def get(self, key: str, /, default: str | _T = None) -> str | _T:
        return self.inner.get(key, default=default)

    def copy(self) -> dict[str, str]:
        return self.inner.copy()


# SAFETY: wrapped to not allow mutation
environ = ReadonlyDict(os.environ)  # noqa: TID251



import pytest
import json

from testlib.fixtures.nix import Nix


@pytest.mark.parametrize(
    ("trusted", "flags", "expected"),
    [
        ([], ["--default-trust"], False),
        ([], ["--force-trusted"], True),
        (["*"], ["--default-trust"], True),
        (["*"], ["--force-untrusted"], False),
    ],
)
def test_trust(nix: Nix, trusted: list[str], flags: list[str], expected: bool):
    nix.settings.add_xp_feature("nix-command", "daemon-trust-override")

    with nix.daemon(flags, settings={"trusted-users": trusted}) as inner:
        trusted = json.loads(inner.nix(["store", "ping", "--json"]).run().ok().stdout)
        assert trusted["trusted"] == expected



from pathlib import Path
from testlib.fixtures.nix import Nix
from testlib.fixtures.file_helper import with_files
from testlib.utils import get_global_asset_pack


@with_files(get_global_asset_pack("simple-drv"))
def test_stdio_forward_failure(nix: Nix, files: Path):
    result = (
        nix.nix_build(
            [
                f"{files}/simple.nix",
                "--builders",
                # build-remote handles the first ssh-ng in-process, but we want to test how nix-daemon
                # behaves in error cases. we need to explicitly set remote-program for the main remote
                # connection since we may choose the wrong binary otherwise. the inner ssh-ng also has
                # a remote-program set to cause predictable failures (rather than using ssh impurely).
                f"ssh-ng://localhost?remote-program={nix.env.dirs.nix_bin_dir}/nix-daemon&remote-store=ssh-ng://localhost?remote-program=false",
                "-j0",
            ]
        )
        .run()
        .expect(1)
    )
    assert "stream ended unexpectedly" in result.stderr_plain
    assert "Lix crashed" not in result.stderr_plain



from pathlib import Path
from testlib.fixtures.nix import Nix
from testlib.fixtures.file_helper import File, with_files


@with_files({"hello": File("hello")})
def test_ssh_relay(nix: Nix):
    """
    test that nesting ssh stores via remote-store url parameters works.
    we don't currently check that each store spawns a new ssh instance,
    but if it didn't we'd almost certainly get remote execution errors.
    """

    nix.settings.add_xp_feature("nix-command")

    ssh_localhost = "ssh://localhost"
    store = ssh_localhost + 3 * f"?remote-store={ssh_localhost}"

    out = nix.nix(["store", "add-path", "--store", store, "hello"]).run().ok().stdout_plain
    assert Path(out).read_text() == "hello"



import stat
import sys
import pytest
import signal
from pathlib import Path
from textwrap import dedent
from testlib.fixtures.env import ManagedEnv
from testlib.fixtures.nix import Nix

COMMANDS = ["copy-closure", "collect-garbage"]


@pytest.fixture(scope="module")
def custom_sub_command_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    return tmp_path_factory.mktemp(__name__)


@pytest.fixture(scope="module")
def failing_sub_command_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    return tmp_path_factory.mktemp(__name__ + "-failing")


@pytest.fixture(scope="module", params=COMMANDS)
def custom_sub_command(request: pytest.FixtureRequest, custom_sub_command_path: Path) -> str:
    # Create an external command that is an alias of an existing one
    lix_cmd = request.param

    executable = custom_sub_command_path / f"lix-{lix_cmd}"
    executable.write_text(
        dedent(f"""\
         #!{sys.executable}
         import os, sys
         # Start with args[0] set to the actual nix command used for testing
         # as we are not making Lix variants of those.
         os.execvp("nix", [ "nix-{lix_cmd}" ] + sys.argv[1:])
         """)
    )
    executable.chmod(stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR)

    return lix_cmd


@pytest.fixture(scope="module")
def failing_sub_command(failing_sub_command_path: Path, custom_sub_command: str) -> str:
    # Create an external command that will intentionally cause an error
    executable = failing_sub_command_path / f"lix-{custom_sub_command}"
    executable.write_text(
        dedent(f"""\
         #!{sys.executable}
         import sys
         sys.exit(42)
         """)
    )
    executable.chmod(stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR)

    return custom_sub_command


@pytest.fixture
def path(custom_sub_command_path: Path, env: ManagedEnv):
    """Provide a modified search path"""
    env.path.append("/some/incorrect")
    env.path.append("/location/for/fun")
    env.path.append(str(custom_sub_command_path))


@pytest.fixture
def path_with_failure(
    request: pytest.FixtureRequest,
    custom_sub_command_path: Path,
    failing_sub_command_path: Path,
    env: ManagedEnv,
):
    """
    Provide a search path with failing binaries inserted in the specified position
    """
    (first, fail) = request.param

    command_path = [str(custom_sub_command_path), str(failing_sub_command_path)]
    if fail:
        command_path.reverse()

    if first:
        env.path.prepend(command_path[1])
        env.path.prepend(command_path[0])
    else:
        env.path.append(command_path[0])
        env.path.append(command_path[1])


@pytest.mark.parametrize(
    ("nix_exe", "flag", "expected"),
    [("nix", False, 1), ("lix", False, 1), ("nix", True, 1), ("lix", True, 0)],
)
@pytest.mark.usefixtures("path")
def test_sub_commands(nix: Nix, custom_sub_command: str, nix_exe: str, flag: bool, expected: int):
    if flag:
        nix.settings.add_xp_feature("lix-custom-sub-commands")

    # Test custom sub commands in various configurations
    nix.nix([custom_sub_command, "--version"], nix_exe=nix_exe).run().expect(expected)


@pytest.mark.parametrize(
    ("path_with_failure", "expected"),
    [((True, True), 42), ((True, False), 0), ((False, True), 42), ((False, False), 0)],
    indirect=["path_with_failure"],
)
@pytest.mark.usefixtures("path_with_failure")
def test_sub_command_path_order(nix: Nix, failing_sub_command: str, expected: int):
    # Test handling of the order of the path for custom sub commands
    # Incidentally also tests passing through exit codes
    nix.settings.add_xp_feature("lix-custom-sub-commands")
    nix.nix([failing_sub_command, "--version"], nix_exe="lix").run().expect(expected)


@pytest.mark.skip(
    reason="TODO(raito): we do not support auto completion for custom sub commands for now"
)
def test_custom_sub_commands_auto_completion(nix: Nix, tmp_path: Path): ...


@pytest.mark.skip(
    reason="TODO(raito): we do not test flag handling for custom sub commands for now"
)
def test_custom_sub_command_flag_handling(nix: Nix, tmp_path: Path):
    # Short, long and multiple flags should be tested as well.
    # `--` special flag
    # test positional arguments, but only `nix-copy-closure` implements some, and it's pesky to test here.
    ...


@pytest.mark.usefixtures("path")
def test_custom_sub_command_signals(nix: Nix, custom_sub_command_path: Path):
    command = custom_sub_command_path / "lix-signals"
    command.write_text(
        dedent(f"""\
            #!{sys.executable}
            import signal
            print(signal.pthread_sigmask(signal.SIG_BLOCK, []))
        """)
    )
    command.chmod(stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR)

    current_mask = signal.pthread_sigmask(signal.SIG_BLOCK, [])
    nix.settings.add_xp_feature("lix-custom-sub-commands")
    result = nix.nix(["signals"], nix_exe="lix").run()
    result.ok()
    assert result.stdout_s.strip() == str(current_mask)



import pytest

from testlib.fixtures.env import ManagedEnv
from testlib.fixtures.file_helper import with_files
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset


@with_files({"trivial.nix": get_global_asset("trivial.nix")})
@pytest.mark.parametrize("prefix", ["by-absolute-path", "by-relative-path"])
def test_nix_path(env: ManagedEnv, nix: Nix, prefix: str):
    env.set_env(
        "NIX_PATH",
        f"non-existent=/non-existent/but-unused-anyways:by-absolute-path={env.dirs.home}:by-relative-path=.",
    )

    expected_path = str(env.dirs.home / "trivial.nix")

    res = (
        nix.nix_instantiate(["--eval", "-E", f"<{prefix}/trivial.nix>", "--restrict-eval"])
        .run()
        .ok()
    )
    assert res.stdout_plain == expected_path
    res = nix.nix_instantiate(["--find-file", f"{prefix}/trivial.nix"]).run().ok()
    path = res.stdout_plain
    assert path == expected_path


@pytest.mark.skip("FIXME(rootile): for some reason the channel cannot be resolved on the builders")
def test_nix_path_nixpkgs(env: ManagedEnv, nix: Nix):
    env.set_env(
        "NIX_PATH",
        f"non-existent=/non-existent/but-unused-anyways:by-absolute-path={env.dirs.home}:by-relative-path=.",
    )
    res = (
        nix.nix_instantiate(
            ["--eval", "-E", "<nixpkgs>", "-I", "nixpkgs=channel:nixos-unstable", "--restrict-eval"]
        )
        .run()
        .ok()
    )
    assert str(env.dirs.nix_store_dir) in res.stdout_plain



import os
import unicodedata
from io import BytesIO
from logging import Logger
from pathlib import Path

import pytest

from testlib.fixtures.nix import Nix, with_diverted_store
from testlib.nar import (
    DirectoryUnordered,
    NarItem,
    NarListener,
    Regular,
    Symlink,
    write_with_export_header,
)

meow_orig = "méow"
meow_nfc_ = unicodedata.normalize("NFC", meow_orig)
meow_nfd_ = unicodedata.normalize("NFD", meow_orig)
meow_nfc = meow_nfc_.encode("utf-8")
meow_nfd = meow_nfd_.encode("utf-8")
assert meow_nfc != meow_nfd

EVIL_NARS: list[tuple[str, NarItem]] = [
    (
        "valid-dir-1",
        DirectoryUnordered(
            [
                (b"a-nested", DirectoryUnordered([(b"loopy", Symlink(b"../abc-nested"))])),
                (b"b-file", Regular(False, b"meow kbity")),
                (b"c-exe", Regular(True, b"#!/usr/bin/env cat\nmeow kbity")),
            ]
        ),
    ),
    (
        "invalid-slashes-1",
        DirectoryUnordered(
            [(b"meow", Symlink(b"meowmeow")), (b"meow/nya", Regular(False, b"eepy"))]
        ),
    ),
    ("invalid-dot-1", DirectoryUnordered([(b".", Symlink(b"meowmeow"))])),
    ("invalid-dot-2", DirectoryUnordered([(b"..", Symlink(b"meowmeow"))])),
    ("invalid-nul-1", DirectoryUnordered([(b"meow\0nya", Symlink(b"meowmeow"))])),
    (
        "invalid-misorder-1",
        DirectoryUnordered(
            [(b"zzz", Regular(False, b"eepy")), (b"kbity", Regular(False, b"meow"))]
        ),
    ),
    (
        "invalid-dupe-1",
        DirectoryUnordered([(b"zzz", Regular(False, b"eepy")), (b"zzz", Regular(False, b"meow"))]),
    ),
    (
        "invalid-dupe-2",
        DirectoryUnordered(
            [
                (b"zzz", DirectoryUnordered([(b"meow", Regular(False, b"kbity"))])),
                (b"zzz", Regular(False, b"meow")),
            ]
        ),
    ),
    (
        "invalid-dupe-3",
        DirectoryUnordered(
            [
                (b"zzz", DirectoryUnordered([(b"meow", Regular(False, b"kbity"))])),
                (b"zzz", DirectoryUnordered([(b"meow", Regular(False, b"kbityy"))])),
            ]
        ),
    ),
    (
        "invalid-dupe-4",
        DirectoryUnordered(
            [
                (b"zzz", Symlink(b"../kbity")),
                (b"zzz", DirectoryUnordered([(b"meow", Regular(False, b"kbityy"))])),
            ]
        ),
    ),
    (
        "invalid-casehack-1",
        DirectoryUnordered(
            [
                (b"ZZZ~nix~case~hack~2", Regular(False, b"meow")),
                (b"zzz~nix~case~hack~1", Regular(False, b"eepy")),
            ]
        ),
    ),
    (
        "invalid-casehack-2",
        DirectoryUnordered(
            [
                (b"ZZZ~nix~case~hack~1", Regular(False, b"meow")),
                (b"zzz~nix~case~hack~1", Regular(False, b"eepy")),
            ]
        ),
    ),
]


@pytest.mark.parametrize(("name", "nar"), EVIL_NARS, ids=next(zip(*EVIL_NARS)))
@with_diverted_store
def test_evil_nar(nix: Nix, name: str, nar: NarItem, logger: Logger):
    bio = BytesIO()

    listener = NarListener(bio)
    write_with_export_header(nar, name.encode(), listener)
    logger.info(nar)

    if name.startswith("valid-"):
        expected_rc = 0
    elif name.startswith("invalid-"):
        expected_rc = 1
    else:
        raise ValueError("bad name", name)

    res = nix.nix_store(["--import"]).with_stdin(bio.getvalue()).run().expect(expected_rc)
    logger.info(res)


def test_unicode_evil_nar(nix: Nix, tmp_path: Path, logger: Logger):
    """
    Depending on the filesystem in use, filenames that are equal modulo unicode
    normalization may hit the same file or not.

    On macOS, such collisions will result in hitting the same file. We detect
    if the fs is like this before checking what Lix does.
    """
    with open(os.path.join(bytes(tmp_path), meow_nfc), "wb") as fh:
        fh.write(b"meow")

    try:
        with open(os.path.join(bytes(tmp_path), meow_nfd), "rb") as fh:
            assert fh.read() == b"meow"
    except FileNotFoundError:
        # normalization is not applied to this system
        pytest.skip("filesystem does not use unicode normalization")

    test_evil_nar(
        nix,
        "invalid-unicode-normalization-1",
        DirectoryUnordered(
            [
                # méow
                (meow_nfd, Regular(False, b"eepy")),
                (meow_nfc, Symlink(b"meowmeow")),
            ]
        ),
        logger,
    )
    test_evil_nar(
        nix,
        "invalid-unicode-normalization-2",
        DirectoryUnordered(
            [
                # méow
                (meow_nfd, Symlink(b"meowmeow")),
                (meow_nfc, Regular(False, b"eepy")),
            ]
        ),
        logger,
    )



import pytest
from testlib.fixtures.nix import Nix


class TestFunctionTrace:
    @pytest.fixture(autouse=True)
    def setup(self, nix: Nix):
        self.nix = nix

    def call(self, expr: str) -> list[str]:
        trace = (
            self.nix.nix_instantiate(["--trace-function-calls", "--expr", expr]).run().stderr_plain
        )
        return [
            " ".join(line.split(" ")[:-1])
            for line in trace.splitlines()
            if line.startswith("function-trace")
        ]

    def test_try_eval(self):
        assert self.call('builtins.tryEval (throw "example")') == [
            "function-trace entered «string»:1:1 at",
            "function-trace entered «string»:1:19 at",
            "function-trace exited «string»:1:19 at",
            "function-trace exited «string»:1:1 at",
        ]

    @pytest.mark.parametrize(
        "expr",
        [
            "({ x }: x) { }",
            '({ x }: x) { x = "x"; y = "y"; }',
            "(x: y: x + y) 1",
            "(x: x) 1 2",
            "1 2",
        ],
    )
    def test_single_call(self, expr: str):
        assert self.call(expr) == [
            "function-trace entered «string»:1:1 at",
            "function-trace exited «string»:1:1 at",
        ]



#!@python@
# ruff: noqa: SIM112 # ignore lowercase env variable names for capnpc as we have them in lower case as arguments

import argparse
import capnp
from pathlib import Path
import os
import subprocess
import sys

if lang := os.environ.get("lix_capnp_lang"):
    outputs = os.environ["lix_capnp_outputs"].split()
    old_cwd = os.environ["lix_capnp_old_cwd"]
    schema = capnp.load("@capnp_include@/capnp/schema.capnp", imports=["@capnp_include@"])
    request = schema.CodeGeneratorRequest.read(sys.stdin)

    subprocess.run([lang], input=request.as_builder().to_bytes()).check_returncode()

    base_dir = Path.cwd()
    os.chdir(old_cwd)

    include = [str(Path(p).resolve()) for p in os.environ["lix_capnp_include"].split(":")]

    if depfile := os.environ["lix_capnp_depfile"]:
        deps = ""
        for input_file in request.requestedFiles:
            deps += " ".join(f"{input_file.filename}.{o}" for o in outputs)
            deps += ":"
            for dep in input_file.imports:
                if dep.name.startswith("/"):
                    for candidate in (Path(i + dep.name) for i in include):
                        if candidate.exists():
                            deps += " " + str(candidate)
                            break
                else:
                    msg = "not handling relative includes"
                    raise RuntimeError(msg)
            deps += "\n\n"
        Path(depfile).write_text(deps)
else:
    parser = argparse.ArgumentParser()
    parser.add_argument("--language")
    parser.add_argument("--outdir")
    parser.add_argument("--src-prefix")
    parser.add_argument("--depfile", default="")
    parser.add_argument("-I", "--include", action="append", default=["@capnp_include@"])
    parser.add_argument("inputs", nargs="+")
    args = parser.parse_args()

    for infile in args.inputs:
        os.environ["lix_capnp_lang"] = f"capnpc-{args.language}"
        os.environ["lix_capnp_include"] = ":".join(args.include)
        os.environ["lix_capnp_depfile"] = args.depfile
        os.environ["lix_capnp_old_cwd"] = str(Path.cwd())
        if args.language == "c++":
            os.environ["lix_capnp_outputs"] = "c++ h"
        else:
            raise RuntimeError("unknown language " + args.language)
        subprocess.run(
            [
                "@capnp@",
                "compile",
                f"-o{sys.argv[0]}:{args.outdir}",
                f"--src-prefix={args.src_prefix}",
                *(f"-I{i}" for i in args.include),
                infile,
            ]
        ).check_returncode()



from typing import Any, Literal, Optional
import re
from pathlib import Path
import json
import dataclasses
import time
from urllib.parse import unquote
import urllib.request
import logging

import requests.auth
import requests
import xdg_base_dirs

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

DEBUG_REQUESTS = False
if DEBUG_REQUESTS:
    urllib3_logger = logging.getLogger('requests.packages.urllib3')
    urllib3_logger.setLevel(logging.DEBUG)
    urllib3_logger.propagate = True

# So, there is a bunch of confusing stuff happening in this file. The gist of why it's Like This is:
#
# nix2container does not concern itself with tags (reasonably enough):
# https://github.com/nlewo/nix2container/issues/59
#
# This is fine. But then we noticed: docker images don't play nice if you have
# multiple architectures you want to abstract over if you don't do special
# things. Those special things are images with manifests containing multiple
# images.
#
# Docker has a data model vaguely analogous to git: you have higher level
# objects referring to a bunch of content-addressed blobs.
#
# A multiarch image is more or less just a manifest that refers to more
# manifests; in OCI it is an Index.
#
# See the API spec here: https://github.com/opencontainers/distribution-spec/blob/v1.0.1/spec.md#definitions
# And the Index spec here: https://github.com/opencontainers/image-spec/blob/v1.0.1/image-index.md
#
# skopeo doesn't *know* how to make multiarch *manifests*:
# https://github.com/containers/skopeo/issues/1136
#
# There is a tool called manifest-tool that is supposed to do this
# (https://github.com/estesp/manifest-tool) but it doesn't support putting in
# annotations on the outer image, and I *really* didn't want to write golang to
# fix that. Thus, a little bit of homebrew containers code.
#
# Essentially what we are doing in here is splatting a bunch of images into the
# registry without tagging them (with a silly workaround to skopeo issues),
# then simply sending a new composite manifest ourselves.

DockerArchitecture = Literal['amd64'] | Literal['arm64']
MANIFEST_MIME = 'application/vnd.oci.image.manifest.v1+json'
INDEX_MIME = 'application/vnd.oci.image.index.v1+json'


@dataclasses.dataclass(frozen=True, order=True)
class ImageMetadata:
    size: int
    digest: str
    """sha256:SOMEHEX"""


@dataclasses.dataclass(frozen=True, order=True)
class OCIIndexItem:
    """Information about an untagged uploaded image."""

    metadata: ImageMetadata

    architecture: DockerArchitecture

    os: str = 'linux'

    def serialize(self):
        return {
            'mediaType': MANIFEST_MIME,
            'size': self.metadata.size,
            'digest': self.metadata.digest,
            'platform': {
                'architecture': self.architecture,
                'os': self.os,
            }
        }


@dataclasses.dataclass(frozen=True)
class OCIIndex:
    manifests: list[OCIIndexItem]

    annotations: dict[str, str]

    def serialize(self):
        return {
            'schemaVersion': 2,
            'manifests': [item.serialize() for item in sorted(self.manifests)],
            'annotations': self.annotations
        }


@dataclasses.dataclass
class TaggingOperation:
    manifest: OCIIndex
    tags: list[str]
    """Tags this image is uploaded under"""


runtime_dir = xdg_base_dirs.xdg_runtime_dir()
config_dir = xdg_base_dirs.xdg_config_home()

AUTH_FILES = ([runtime_dir / 'containers/auth.json'] if runtime_dir else []) + \
    [config_dir / 'containers/auth.json', Path.home() / '.docker/config.json']


# Copied from Werkzeug https://github.com/pallets/werkzeug/blob/62e3ea45846d06576199a2f8470be7fe44c867c1/src/werkzeug/http.py#L300-L325
def parse_list_header(value: str) -> list[str]:
    """Parse a header value that consists of a list of comma separated items according
    to `RFC 9110 <https://httpwg.org/specs/rfc9110.html#abnf.extension>`__.

    This extends :func:`urllib.request.parse_http_list` to remove surrounding quotes
    from values.

    .. code-block:: python

        parse_list_header('token, "quoted value"')
        ['token', 'quoted value']

    This is the reverse of :func:`dump_header`.

    :param value: The header value to parse.
    """
    result = []

    for item in urllib.request.parse_http_list(value):
        if len(item) >= 2 and item[0] == item[-1] == '"':
            item = item[1:-1]

        result.append(item)

    return result


# https://www.rfc-editor.org/rfc/rfc2231#section-4
_charset_value_re = re.compile(
    r"""
    ([\w!#$%&*+\-.^`|~]*)'  # charset part, could be empty
    [\w!#$%&*+\-.^`|~]*'  # don't care about language part, usually empty
    ([\w!#$%&'*+\-.^`|~]+)  # one or more token chars with percent encoding
    """,
    re.ASCII | re.VERBOSE,
)


# Copied from: https://github.com/pallets/werkzeug/blob/62e3ea45846d06576199a2f8470be7fe44c867c1/src/werkzeug/http.py#L327-L394
def parse_dict_header(value: str) -> dict[str, str | None]:
    """Parse a list header using :func:`parse_list_header`, then parse each item as a
    ``key=value`` pair.

    .. code-block:: python

        parse_dict_header('a=b, c="d, e", f')
        {"a": "b", "c": "d, e", "f": None}

    This is the reverse of :func:`dump_header`.

    If a key does not have a value, it is ``None``.

    This handles charsets for values as described in
    `RFC 2231 <https://www.rfc-editor.org/rfc/rfc2231#section-3>`__. Only ASCII, UTF-8,
    and ISO-8859-1 charsets are accepted, otherwise the value remains quoted.

    :param value: The header value to parse.

    .. versionchanged:: 3.0
        Passing bytes is not supported.

    .. versionchanged:: 3.0
        The ``cls`` argument is removed.

    .. versionchanged:: 2.3
        Added support for ``key*=charset''value`` encoded items.

    .. versionchanged:: 0.9
       The ``cls`` argument was added.
    """
    result: dict[str, str | None] = {}

    for item in parse_list_header(value):
        key, has_value, value = item.partition("=")
        key = key.strip()

        if not has_value:
            result[key] = None
            continue

        value = value.strip()
        encoding: str | None = None

        if key[-1] == "*":
            # key*=charset''value becomes key=value, where value is percent encoded
            # adapted from parse_options_header, without the continuation handling
            key = key[:-1]
            match = _charset_value_re.match(value)

            if match:
                # If there is a charset marker in the value, split it off.
                encoding, value = match.groups()
                assert encoding
                encoding = encoding.lower()

            # A safe list of encodings. Modern clients should only send ASCII or UTF-8.
            # This list will not be extended further. An invalid encoding will leave the
            # value quoted.
            if encoding in {"ascii", "us-ascii", "utf-8", "iso-8859-1"}:
                # invalid bytes are replaced during unquoting
                value = unquote(value, encoding=encoding)

        if len(value) >= 2 and value[0] == value[-1] == '"':
            value = value[1:-1]

        result[key] = value

    return result


def parse_www_authenticate(www_authenticate):
    scheme, _, rest = www_authenticate.partition(' ')
    scheme = scheme.lower()
    rest = rest.strip()

    parsed = parse_dict_header(rest.rstrip('='))
    return parsed


class AuthState:

    def __init__(self, auth_files: list[Path] = AUTH_FILES):
        self.auth_map: dict[str, str] = {}
        for f in auth_files:
            self.auth_map.update(AuthState.load_auth_file(f))
        self.token_cache: dict[str, str] = {}

    @staticmethod
    def load_auth_file(path: Path) -> dict[str, str]:
        if path.exists():
            with path.open() as fh:
                try:
                    json_obj = json.load(fh)
                    auths = json_obj.get('auths', {})
                    return {k: v['auth'] for k, v in auths.items()}
                except (json.JSONDecodeError, KeyError) as e:
                    log.exception('JSON decode error in %s', path, exc_info=e)
        return {}

    def get_token(self, hostname: str) -> Optional[str]:
        return self.token_cache.get(hostname)

    def obtain_token(self, session: requests.Session, token_endpoint: str,
                     scope: str, service: str, image_path: str) -> str:
        authority, _, _ = image_path.partition('/')
        if tok := self.get_token(authority):
            return tok

        creds = self.find_credential_for(image_path)
        if not creds:
            raise ValueError('No credentials available for ' + image_path)

        resp = session.get(token_endpoint,
                           params={
                               'client_id': 'lix-releng',
                               'scope': scope,
                               'service': service,
                           },
                           headers={
                               'Authorization': 'Basic ' + creds
                           }).json()
        token = resp['token']
        self.token_cache[authority] = token
        return token

    def find_credential_for(self, image_path: str):
        trails = image_path.split('/')
        for i in range(len(trails)):
            prefix = '/'.join(trails[:len(trails) - i])
            if prefix in self.auth_map:
                return self.auth_map[prefix]

        return None


class RegistryAuthenticator(requests.auth.AuthBase):
    """Authenticates to an OCI compliant registry"""

    def __init__(self, auth_state: AuthState, session: requests.Session,
                 image: str):
        self.auth_map: dict[str, str] = {}
        self.image = image
        self.session = session
        self.auth_state = auth_state

    def response_hook(self, r: requests.Response,
                      **kwargs: Any) -> requests.Response:
        if r.status_code == 401:
            www_authenticate = r.headers.get('www-authenticate', '').lower()
            parsed = parse_www_authenticate(www_authenticate)
            assert parsed

            tok = self.auth_state.obtain_token(
                self.session,
                parsed['realm'],  # type: ignore
                parsed['scope'],  # type: ignore
                parsed['service'],  # type: ignore
                self.image)

            new_req = r.request.copy()
            new_req.headers['Authorization'] = 'Bearer ' + tok

            return self.session.send(new_req)
        else:
            return r

    def __call__(self,
                 r: requests.PreparedRequest) -> requests.PreparedRequest:
        authority, _, _ = self.image.partition('/')
        auth_may = self.auth_state.get_token(authority)

        if auth_may:
            r.headers['Authorization'] = 'Bearer ' + auth_may

        r.register_hook('response', self.response_hook)
        return r


class Registry:

    def __init__(self, session: requests.Session):
        self.auth_state = AuthState()
        self.session = session

    def image_info(self, image_path: str, manifest_id: str) -> ImageMetadata:
        authority, _, path = image_path.partition('/')
        resp = self.session.head(
            f'https://{authority}/v2/{path}/manifests/{manifest_id}',
            headers={'Accept': MANIFEST_MIME},
            auth=RegistryAuthenticator(self.auth_state, self.session,
                                       image_path))
        resp.raise_for_status()
        return ImageMetadata(int(resp.headers['content-length']),
                             resp.headers['docker-content-digest'])

    def delete_tag(self, image_path: str, tag: str):
        authority, _, path = image_path.partition('/')
        resp = self.session.delete(
            f'https://{authority}/v2/{path}/manifests/{tag}',
            headers={'Content-Type': INDEX_MIME},
            auth=RegistryAuthenticator(self.auth_state, self.session,
                                       image_path))
        resp.raise_for_status()

    def _upload_index(self, image_path: str, tag: str, index: OCIIndex):
        authority, _, path = image_path.partition('/')
        body = json.dumps(index.serialize(),
                          separators=(',', ':'),
                          sort_keys=True)

        resp = self.session.put(
            f'https://{authority}/v2/{path}/manifests/{tag}',
            data=body,
            headers={'Content-Type': INDEX_MIME},
            auth=RegistryAuthenticator(self.auth_state, self.session,
                                       image_path))
        resp.raise_for_status()

        return resp.headers['Location']

    def upload_index(self,
                     image_path: str,
                     tag: str,
                     index: OCIIndex,
                     retries=20,
                     retry_delay=1):
        # eventual consistency lmao
        for _ in range(retries):
            try:
                return self._upload_index(image_path, tag, index)
            except requests.HTTPError as e:
                if e.response.status_code != 404:
                    raise

            time.sleep(retry_delay)



import logging
import stat
from collections.abc import Callable
from subprocess import CalledProcessError

import pytest
from _pytest.logging import LogCaptureFixture
from testlib.fixtures.command import Command
from testlib.fixtures.env import ManagedEnv
from testlib.fixtures.file_helper import File, with_files


def test_command_valid_runs(command: Callable[[list[str]], Command]):
    cmd = command(["echo", "water"])
    cmd.run().ok()


def test_command_captures_stdout(command: Callable[[list[str]], Command]):
    cmd = command(["echo", "fire"])
    res = cmd.run().ok()
    assert res.stdout_s == "fire\n"


def test_command_plain_strips(command: Callable[[list[str]], Command]):
    cmd = command(["echo", "   earth   "])
    res = cmd.run().ok()
    assert res.stdout_plain == "earth"


def test_command_stdin_passed_correctly(command: Callable[[list[str]], Command]):
    inp = b"air"
    cmd = command(["cat", "/dev/stdin"]).with_stdin(inp)
    res = cmd.run().ok()
    assert res.stdout == inp


def test_command_expect_failure(env: ManagedEnv):
    env.path.add_program("grep")
    Command(_env=env, argv=["grep", "xxx"], stdin=b"").run().expect(1)


@with_files(
    {
        "script.sh": File(
            "#!/bin/sh\necho forb\nexit 1", mode=stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO
        )
    }
)
def test_command_ok_fails_on_bad_exit_code(
    command: Callable[[list[str]], Command], caplog: LogCaptureFixture
):
    cmd = command(["./script.sh"])

    with pytest.raises(CalledProcessError), caplog.at_level(logging.ERROR):
        cmd.run().ok()
    msgs = caplog.messages
    assert len(msgs) == 2
    out_msg, err_msg = msgs
    assert out_msg == "stdout: forb\n"
    assert err_msg == "stderr: "


@with_files(
    {
        "script.sh": File(
            "#!/bin/sh\necho drgn fops\nexit 2", mode=stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO
        )
    }
)
def test_command_exec_fails_on_other_bad_exit_code(
    command: Callable[[list[str]], Command], caplog: LogCaptureFixture
):
    cmd = command(["./script.sh"])

    with pytest.raises(CalledProcessError), caplog.at_level(logging.ERROR):
        cmd.run().expect(1)
    msgs = caplog.messages
    assert len(msgs) == 2
    out_msg, err_msg = msgs
    assert out_msg == "stdout: drgn fops\n"
    assert err_msg == "stderr: "



import argparse
import re
from collections.abc import Callable
from pathlib import Path
from typing import Any

import frontmatter


def cxx_escape_character(c: str) -> str:
    if 0x20 <= ord(c) < 0x7F and c != '"' and c != "?" and c != "\\":
        return c
    if c == "\t":
        return r"\t"
    if c == "\n":
        return r"\n"
    if c == "\r":
        return r"\r"
    if c == '"':
        return r"\""
    if c == "?":
        return r"\?"
    if c == "\\":
        return r"\\"
    if ord(c) <= 0xFFFF:
        return str.format(r"\u{:04x}", ord(c))
    return str.format(r"\U{:08x}", ord(c))


def cxx_literal(v: Any) -> str:
    if v is None:
        return "std::nullopt"
    if v is False:
        return "false"
    if v is True:
        return "true"
    if isinstance(v, int):
        return str(v)
    if isinstance(v, str):
        return "".join(['"', *(cxx_escape_character(c) for c in v), '"'])
    if isinstance(v, list):
        return f"{{{', '.join([cxx_literal(item) for item in v])}}}"
    msg = f"cannot represent {v!r} in C++"
    raise NotImplementedError(msg)


def get_experimental_features(
    base_path: str, human_names: list[str | None]
) -> dict[str | None, str]:
    experimental_feature_files = {
        f"{base_path}/{xp_name}.md" for xp_name in human_names if xp_name is not None
    }

    from build_extra_features import ExtraFeature  # noqa: PLC0415 # Avoid cyclic import

    experimental_features_data = load_data(list(experimental_feature_files), ExtraFeature)
    experimental_features: dict[str | None, str] = {
        xf.name: f"Xp::{xf.internal_name}" for _, xf in experimental_features_data
    }
    experimental_features[None] = "std::nullopt"
    return experimental_features


FIELD_RENAMES = {"type": "type_str", "content": "documentation"}


def load_data[T](defs: list[str], parse_function: type[T]) -> list[tuple[str, T]]:
    data = []
    for path in defs:
        try:
            datum = {
                # convert camelCase to snake_case
                re.sub(r"(?<=.)([A-Z])", lambda m: f"_{m.group(1).lower()}", k): v
                for k, v in frontmatter.load(path).to_dict().items()
            }

            for post_name, field_name in FIELD_RENAMES.items():
                if post_name in datum:
                    datum[field_name] = datum.pop(post_name)

            data.append((path, parse_function(**datum)))
        except Exception as e:
            e.add_note(f"in {path}")
            raise
    return data


def generate_file[T](
    path: str | None,
    data: list[T],
    sort_key_function: Callable[[T], str],
    generate_function: Callable[[T], str],
):
    if path is not None:
        with Path(path).open("w") as out:
            for path, datum in sorted(
                data, key=lambda path_and_datum: sort_key_function(path_and_datum[1])
            ):
                try:
                    text = generate_function(datum)
                    out.write(text)
                except Exception as e:
                    e.add_note(f"in {path}")
                    raise


def get_argument_parser() -> argparse.ArgumentParser:
    ap = argparse.ArgumentParser()
    ap.add_argument("--header", help="Path of the header to generate")
    ap.add_argument("--docs", help="Path of the documentation file to generate")
    ap.add_argument("defs", help="Builtin definition files", nargs="+")

    return ap



import pytest
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset
from testlib.fixtures.file_helper import with_files, File


drv = """
  builtins.derivation {
    name = "foo";
    builder = /bin/sh;
    system = builtins.currentSystem;
    requiredSystemFeatures = [ "glitter" ];
  }
"""

local_drv = """
  with import ./config.nix; mkDerivation {
    name = "foo";
    buildCommand = ''
      touch $out
    '';
    preferLocalBuild = true;
    requiredSystemFeatures = [ "glitter" ];
  }
"""


@pytest.fixture(autouse=True)
def set_features(nix: Nix):
    nix.settings.system_features = "glitter"


def test_j0_without_remotes_fails(nix: Nix):
    result = nix.nix_build(["-j0", "--expr", drv, "--builders", ""]).run().expect(1)
    assert (
        "error: unable to start any build; either set '--max-jobs' to a non-zero value or enable remote builds."
        in result.stderr_s
    )


def test_j0_with_mismatched_remotes_fails(nix: Nix):
    remote = f"ssh://localhost?remote-store={nix.env.dirs.home}/machine1"
    result = nix.nix_build(["-j0", "--expr", drv, "--builders", remote]).run().expect(1)
    assert (
        "error: unable to start any build; remote machines may not have all required system features."
        in result.stderr_s
    )


@with_files({"config.nix": get_global_asset("config.nix"), "default.nix": File(local_drv)})
def test_j0_without_local_jobs(nix: Nix):
    result = nix.nix_build(["-j0", "--extra-local-jobs", "0"]).run().expect(1)
    assert (
        "error: unable to start any build; either set '--max-jobs' to a non-zero value or enable remote builds."
        in result.stderr_s
    )


@with_files({"config.nix": get_global_asset("config.nix"), "default.nix": File(local_drv)})
def test_j0_with_local_jobs(nix: Nix):
    nix.nix_build(["-j0"]).run().ok()


# TODO(Raito): test -j1 --extra-local-jobs=1 in another file and ensure that preferLocalBuild *AND* another build job can run properly at the same time.



#!@python@

# ruff: noqa: T201 # Our generated MD file is expected on stdout
# This is the only file and is executable

from collections import defaultdict
import frontmatter
import pathlib
import textwrap
from typing import Any
import dataclasses
import yaml
import argparse


GH_ROOT = "https://github.com/"
GH_REPO_BASE = "https://github.com/NixOS/nix"
FORGEJO_REPO_BASE = "https://git.lix.systems/lix-project/lix"
FORGEJO_ROOT = "https://git.lix.systems/"
GERRIT_BASE = "https://gerrit.lix.systems/c/lix/+"
KNOWN_KEYS = ("synopsis", "cls", "issues", "prs", "significance", "category", "credits")

SIGNIFICANCECES = {None: 0, "significant": 10}

# This is just hardcoded for better validation. If you think there should be
# more of them, feel free to add more.
#
# Please update doc/manual/src/contributing/hacking.md if you do. Thanks~
CATEGORIES = [
    "Breaking Changes",
    "Features",
    "Improvements",
    "Fixes",
    "Packaging",
    "Development",
    "Miscellany",
]


@dataclasses.dataclass
class AuthorInfo:
    name: str
    github: str | None = None
    forgejo: str | None = None
    display_name: str | None = None

    def show_name(self) -> str:
        return self.display_name or self.name

    def __str__(self) -> str:
        if self.forgejo:
            return f"[{self.show_name()}]({FORGEJO_ROOT}{self.forgejo})"
        if self.github:
            return f"[{self.show_name()}]({GH_ROOT}{self.github})"
        return self.show_name()


class AuthorInfoDB:
    def __init__(self, author_info: dict[str, dict], throw_on_missing: bool):
        self.author_info = {name: AuthorInfo(name=name, **d) for (name, d) in author_info.items()}
        self.throw_on_missing = throw_on_missing

    def __getitem__(self, name: str) -> str:
        if name in self.author_info:
            return str(self.author_info[name])
        if self.throw_on_missing:
            msg = f"Missing author info for author {name}"
            raise Exception(msg)
        return name


def format_link(ident: str, gh_part: str, fj_part: str) -> str:
    if ident.isdigit():
        ident = f"lix#{ident}"

    # Unify the strings, in order for them to be consistent throughout the release notes.
    ident = ident.replace("gh#", "nix#").replace("fj#", "lix#")

    if ident.startswith("nix#"):
        num, link, base = int(ident[4:]), ident, f"{GH_REPO_BASE}/{gh_part}"
    elif ident.startswith("lix#"):
        num, link, base = int(ident[4:]), ident, f"{FORGEJO_REPO_BASE}/{fj_part}"
    else:
        msg = f"unrecognized reference format: {ident}"
        raise ValueError(msg)
    return f"[{link}]({base}/{num})"


def format_issue(issue: str) -> str:
    return format_link(issue, "issues", "issues")


def format_pr(pr: str) -> str:
    return format_link(pr, "pull", "pulls")


def format_cl(clid: int) -> str:
    return f"[cl/{clid}]({GERRIT_BASE}/{clid})"


def plural_list(strs: list[str]) -> str:
    if len(strs) <= 1:
        return "".join(strs)
    comma = "," if len(strs) >= 3 else ""
    return "{}{} and {}".format(", ".join(strs[:-1]), comma, strs[-1])


def listify(li: list | int) -> list:
    if not isinstance(li, list):
        return [li]
    return li


def do_category(author_info: AuthorInfoDB, entries: list[tuple[pathlib.Path, Any]]):
    for p, entry in sorted(
        entries, key=lambda e: (-SIGNIFICANCECES[e[1].metadata.get("significance")], e[0])
    ):
        try:
            header = entry.metadata["synopsis"]
            links = []
            links += [format_issue(str(s)) for s in listify(entry.metadata.get("issues", []))]
            links += [format_pr(str(s)) for s in listify(entry.metadata.get("prs", []))]
            links += [format_cl(int(cl)) for cl in listify(entry.metadata.get("cls", []))]
            if links != []:
                header += " " + " ".join(links)

            if header:
                print(f"- {header}")
                print()
            else:
                print("- ", end="")

            print(textwrap.indent(entry.content, "  "))
            if credits_authors := listify(entry.metadata.get("credits", [])):
                print()
                print(
                    textwrap.indent(
                        f"Many thanks to {plural_list([author_info[c] for c in credits_authors])} for this.",
                        "  ",
                    )
                )

            # Blank line after each entry.
            print()
        except Exception as e:
            e.add_note(f"in {p}")
            raise


def run_on_dir(author_info: AuthorInfoDB, d: str):
    d = pathlib.Path(d)
    if not d.is_dir():
        msg = f"provided path {d} is not a directory"
        raise ValueError(msg)
    paths = d.glob("[!.]*.md")
    entries = defaultdict(list)
    for p in paths:
        try:
            e = frontmatter.load(p)  # type: ignore
            if "synopsis" not in e.metadata:
                msg = "missing synopsis"
                raise ValueError(msg)
            unknown_keys = set(e.metadata.keys()) - set(KNOWN_KEYS)
            if unknown_keys:
                msg = f"unknown keys: {unknown_keys}"
                raise ValueError(msg)
            category = e.metadata.get("category", "Miscellany")
            if category not in CATEGORIES:
                msg = f"unknown category: {category}"
                raise ValueError(msg)
            entries[category].append((p, e))
        except Exception as e:
            e.add_note(f"in {p}")
            raise

    for category in CATEGORIES:
        if entries[category]:
            print("##", category)
            # Blank line after each heading.
            print()
            do_category(author_info, entries[category])
            # Blank line after each category.
            # This turns into two blank lines when combined with the blank line
            # after each entry.
            print()


def main():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        "--change-authors",
        help="File name of the change authors metadata YAML file",
        type=argparse.FileType("r"),
    )
    ap.add_argument("dirs", help="Directories to run on", nargs="+")

    args = ap.parse_args()

    author_info = (
        AuthorInfoDB(yaml.safe_load(args.change_authors), throw_on_missing=True)
        if args.change_authors
        else AuthorInfoDB({}, throw_on_missing=False)
    )

    for d in args.dirs:
        run_on_dir(author_info, d)


if __name__ == "__main__":
    main()



from collections.abc import Callable
from pathlib import Path

from lang.test_lang import test_eval_okay as nix_eval
from testlib.fixtures.file_helper import with_files, CopyFile, Symlink, AssetSymlink, File
from testlib.fixtures.nix import Nix
from testlib.fixtures.snapshot import Snapshot


@with_files(
    {
        "readDir": {
            "foo": {},
            "ldir": Symlink("./foo"),
            "bar": File(""),
            "linked": Symlink("./bar"),
        },
        "in.nix": CopyFile("in.nix"),
        "out.exp": AssetSymlink("eval-okay.out.exp"),
        "err.exp": AssetSymlink("eval-okay.err.exp"),
    }
)
def test_read_file_type(files: Path, nix: Nix, snapshot: Callable[[str], Snapshot]):
    nix_eval(files, nix, [], snapshot)



from testlib.utils import FileDeclaration, get_global_asset_pack, CopyTemplate, File
from testlib.environ import environ


def simple_flake() -> FileDeclaration:
    return get_global_asset_pack("simple-drv") | {
        "flake.nix": CopyTemplate("assets/simple-flake.nix", {"system": environ.get("system")})
    }


def dependent_flake() -> FileDeclaration:
    return {
        "flake.nix": File(f"""{{
          outputs = {{ self, flake1 }}: {{
            packages.{environ.get("system")}.default = flake1.packages.{environ.get("system")}.default;
            expr = assert builtins.pathExists ./flake.lock; 123;
          }};
        }}""")
    }



from logging import Logger
from pathlib import Path
from textwrap import dedent
from testlib.fixtures.nix import Nix
import re


def test_purity_traversal(nix: Nix, tmp_path: Path, logger: Logger):
    error_re = re.compile(r"error: access to absolute path '.+' is forbidden in pure eval mode")

    flake_dir = tmp_path / "flake"
    flake_dir.mkdir()

    evilpath = tmp_path / "sekrit.txt"
    evilpath.write_text("kitty kitty")

    evilnix = tmp_path / "default.nix"
    evilnix.write_text('"woof"')

    goodnix = flake_dir / "good.nix"
    goodnix.write_text("1")

    nested_dir = flake_dir / "nested" / "nested2"
    nested_dir.mkdir(parents=True)
    (nested_dir / "good.nix").write_text("1")

    (flake_dir / "evil-link").symlink_to(evilpath)
    (flake_dir / "evil-default-nix").symlink_to(tmp_path)
    (flake_dir / "less-evil-link").symlink_to(tmp_path / "link-to-flake")
    (tmp_path / "link-to-flake").symlink_to(flake_dir / "flake.nix")

    (flake_dir / "nested-good.nix").symlink_to("nested/nested2/good.nix")
    (flake_dir / "nested-bad.nix").symlink_to("nested/../../nested2/good.nix")

    (flake_dir / "flake.nix").write_text(
        dedent("""
        {
            inputs = {};
            outputs = inputs: {
                bad1 = "${@ABSPATH@}";
                bad2 = builtins.readFile "${@ABSPATH@}";
                bad3 = builtins.readFile @ABSPATH@;
                bad4 = builtins.readFile ./evil-link;
                bad5 = builtins.readFile "${./evil-link}";
                bad6 = builtins.readFile ./less-evil-link;
                bad7 = builtins.readFile "${./less-evil-link}";
                bad8 = import ./evil-default-nix;
                bad9 = import ./nested-bad.nix;

                good1 = builtins.readFile "${inputs.self.outPath}/good.nix";
                good2 = builtins.readFile ./good.nix;
                good3 = builtins.readFile "${./good.nix}";
                good4 = toString (import ./good.nix);
                good5 = toString (import ./nested-good.nix);
            };
        }
    """).replace("@ABSPATH@", str(evilpath.absolute()))
    )

    for idx in range(1, 10):
        cmd = nix.nix(["eval", f".#bad{idx}"], flake=True)
        cmd.cwd = flake_dir
        res = cmd.run().expect(1)
        logger.info(res.stderr_plain)
        assert error_re.search(res.stderr_plain)
    for idx in range(1, 6):
        cmd = nix.nix(["eval", f".#good{idx}"], flake=True)
        cmd.cwd = flake_dir
        res = cmd.run().expect(0)
        assert res.stdout_plain == '"1"'



import pytest
import re
import textwrap
from urllib.parse import urlencode, quote
import sys

from testlib.fixtures.nix import Nix
from testlib.fixtures.file_helper import with_files
from testlib.utils import get_global_asset
from testlib.fixtures.env import ManagedEnv


@pytest.fixture
def busybox_args(env: ManagedEnv) -> list[str]:
    return ["--arg", "busybox", env.path.which("busybox")]


@pytest.fixture(autouse=True)
def _setup_for_remote_builds(nix: Nix, env: ManagedEnv):
    # always add bash, otherwise lix can't execute the build hook
    env.path.add_program("bash")
    # we don't always use this feature, but it also doesn't hurt
    nix.settings.add_xp_feature("daemon-trust-override")


def _builders(proto: str, flags: list[str], env: ManagedEnv) -> str:
    prog = "nix-store" if proto == "ssh" else "nix-daemon"
    script = f"""\
        #!{sys.executable}
        import os, sys
        os.execvp("{env.dirs.nix_bin_dir}/nix", [*{[prog, *flags]!s}, *sys.argv[1:]])
    """
    path = env.dirs.test_root / "remote-builder" / "launch.py"
    path.parent.mkdir()
    path.write_text(textwrap.dedent(script))
    path.chmod(0o755)

    remote_store = "local?" + urlencode(
        {"system-features": "foo bar baz", "root": str(env.dirs.home / "remote")}, quote_via=quote
    )
    uri_args = urlencode(
        {"remote-program": str(path), "remote-store": remote_store}, quote_via=quote
    )
    return textwrap.dedent(f"""
        version = 1

        [machines.remote]
        uri = "{proto}://localhost?{uri_args}"
        jobs = 8
        speed-factor = 1
        supported-features = [ "foo", "bar", "baz" ]
    """)


@pytest.mark.full_sandbox
@with_files(
    {
        "build-hook.nix": get_global_asset("build-hook.nix"),
        "config.nix": get_global_asset("config.nix"),
    }
)
def test_remote_trustless_unsigned(nix: Nix, env: ManagedEnv, busybox_args: list[str]):
    # We first build a dependency of the derivation we eventually want to build.
    nix.nix_build(
        [
            "build-hook.nix",
            "-A",
            "passthru.input2",
            *busybox_args,
            "--option",
            "system-features",
            "bar",
        ]
    ).run().ok()

    # Now when we go to build that downstream derivation, Lix will try to
    # copy our already-build `input2` to the remote store. That store object
    # is input-addressed, so this will fail.

    result = nix.nix_build(
        [
            "build-hook.nix",
            "--max-jobs",
            "0",
            *busybox_args,
            "--builders",
            _builders("ssh-ng", ["--force-untrusted"], env),
        ]
    ).run()
    result.expect(1)
    assert re.findall(
        r"cannot add path '[^ ]*' because it lacks a signature by a trusted key",
        result.stderr_plain,
    )


@pytest.mark.full_sandbox
@pytest.mark.parametrize(
    ("protocol", "flags"), [("ssh", []), ("ssh-ng", []), ("ssh-ng", ["--force-untrusted"])]
)
@with_files(
    {
        "build-hook.nix": get_global_asset("build-hook.nix"),
        "config.nix": get_global_asset("config.nix"),
    }
)
def test_remote_trustless_ia(
    nix: Nix, env: ManagedEnv, busybox_args: list[str], protocol: str, flags: list[str]
):
    result = nix.nix_build(
        [
            "build-hook.nix",
            "--max-jobs",
            "0",
            *busybox_args,
            "--builders",
            _builders(protocol, flags, env),
        ]
    ).run()
    result.ok()

    out_path = (env.dirs.home / "result").readlink()
    assert nix.physical_store_path_for(out_path).read_text() == "FOO BAR BAZ\n"


@pytest.mark.full_sandbox
@pytest.mark.parametrize(("protocol", "flags"), [("ssh", []), ("ssh-ng", ["--force-untrusted"])])
@with_files(
    {
        "build-hook-ca-fixed.nix": get_global_asset("build-hook-ca-fixed.nix"),
        "config.nix": get_global_asset("config.nix"),
    }
)
def test_remote_trustless_ca(
    nix: Nix, env: ManagedEnv, busybox_args: list[str], protocol: str, flags: list[str]
):
    # Remote doesn't trusts us, but this is fine because we are only
    # building (fixed) CA derivations.
    result = nix.nix_build(
        [
            "build-hook-ca-fixed.nix",
            "--max-jobs",
            "0",
            *busybox_args,
            "--builders",
            _builders(protocol, flags, env),
        ]
    ).run()
    result.ok()

    out_path = (env.dirs.home / "result").readlink()
    assert nix.physical_store_path_for(out_path).read_text() == "FOO BAR BAZ\n"



from pathlib import Path
import subprocess
import datetime

from .version import MAJOR, VERSION, RELEASE_NAME

MANUAL = Path('doc/manual')
RELEASE_NOTES_BASE = MANUAL / 'src/release-notes'
VERSION_RL = RELEASE_NOTES_BASE / f'rl-{MAJOR}.md'
SUMMARY = MANUAL / 'src/SUMMARY.md'

def add_to_summary(date: str):
    # N.B: This kind of duplicates gitutils.is_maintenance_branch, but it's a more clear
    # check that allows potentially releasing a -rc without release notes being
    # moved, then in .0 actually move the release notes in place.
    if VERSION_RL.exists():
        return

    MARKER = '  <!-- RELENG-AUTO-INSERTION-MARKER'

    new_lines = []
    for line in SUMMARY.read_text().splitlines():
        new_lines.append(line)
        if MARKER in line:
            indent, _, _ = line.partition(MARKER)
            new_lines.append(f'{indent}- [Lix {MAJOR} ({date})](release-notes/rl-{MAJOR}.md)')

    # make pre-commit happy about one newline
    text = '\n'.join(new_lines).rstrip()
    text += '\n'
    SUMMARY.write_text(text)

def build_release_notes_to_file():
    date = datetime.datetime.now().strftime('%Y-%m-%d')
    add_to_summary(date)

    print('[+] Preparing release notes')
    RELEASE_NOTES_PATH = Path('doc/manual/rl-next')

    if RELEASE_NOTES_PATH.is_dir():
        notes_body = subprocess.check_output(['build-release-notes', '--change-authors', 'doc/manual/change-authors.yml', RELEASE_NOTES_PATH]).decode()
    else:
        # I guess nobody put release notes on their changes?
        print('[-] Warning: seemingly missing any release notes, not worrying about it')
        notes_body = ''

    rl_path = Path(RELEASE_NOTES_BASE / f'rl-{MAJOR}.md')

    existing_rl = ''
    try:
        with open(rl_path, 'r') as fh:
            existing_rl = fh.read()
    except FileNotFoundError:
        pass

    minor_header = f'# Lix {VERSION} ({date})'

    header = f'# Lix {MAJOR} "{RELEASE_NAME}"'
    if existing_rl.startswith(header):
        # strip the header off for minor releases
        lines = existing_rl.splitlines()
        header = lines[0]
        existing_rl = '\n'.join(lines[1:])
    else:
        header += f' ({date})\n\n'

    header += '\n' + minor_header + '\n'

    notes = header
    notes += notes_body
    notes += "\n\n"
    notes += existing_rl

    # make pre-commit happy about one newline
    notes = notes.rstrip()
    notes += "\n"

    with open(rl_path, 'w') as fh:
        fh.write(notes)

    return rl_path



from testlib.fixtures.nix import Nix
from testlib.fixtures.command import Command, CommandResult
from testlib.fixtures.file_helper import with_files
from testlib.fixtures.env import ManagedEnv
from testlib.environ import environ
from pathlib import Path
import pytest
from .common import simple_flake, dependent_flake
from collections.abc import Callable
import glob

system = environ.get("system")

files = {"flake-hg1": simple_flake(), "flake-hg2": dependent_flake()}

type Hg = Callable[..., CommandResult]


@pytest.fixture
def hg(env: ManagedEnv, command: Callable[..., Command]) -> Hg:
    env.path.add_program("hg")

    def wrap(*args, **kwargs) -> CommandResult:
        return command(["hg", *args], **kwargs).run().ok()

    return wrap


@pytest.fixture(autouse=True)
def common_init(nix: Nix, files: Path, hg: Hg):
    nix.settings.add_xp_feature("nix-command", "flakes")
    nix.settings.flake_registry = str(nix.env.dirs.test_root / "registry.json")

    hg("init", files / "flake-hg1")
    nix.nix(
        [
            "registry",
            "add",
            "--registry",
            nix.settings.flake_registry,
            "flake1",
            f"hg+file://{files}/flake-hg1",
        ]
    ).run().ok()

    hg("init", files / "flake-hg2")

    hg("add", *glob.glob(str(files / "flake-hg1/*")))  # noqa PTH207: Path.glob *does the wrong thing*
    hg(
        "commit",
        "--config",
        "ui.username=foobar@example.org",
        files / "flake-hg1",
        "-m",
        "Initial commit",
    )

    hg("add", files / "flake-hg2/flake.nix")
    hg(
        "commit",
        "--config",
        "ui.username=foobar@example.org",
        files / "flake-hg2",
        "-m",
        "Initial commit",
    )


@with_files(files)
class TestFlakeHg:
    def test_build(self, nix: Nix, files: Path):
        nix.nix(["build", f"hg+file://{files}/flake-hg2"]).run().ok()
        assert (nix.env.dirs.home / "result/hello").exists()

    def test_metadata_dirty(self, nix: Nix, files: Path):
        metadata = (
            nix.nix(["flake", "metadata", "--json", f"hg+file://{files}/flake-hg2"]).run().json()
        )
        assert "revision" not in metadata

    def test_eval(self, nix: Nix, files: Path):
        result = nix.nix(["eval", f"hg+file://{files}/flake-hg2#expr"]).run().ok()
        assert result.stdout_s == "123\n"

    def test_eval_no_dirty(self, nix: Nix, files: Path):
        result = (
            nix.nix(["eval", f"hg+file://{files}/flake-hg2#expr", "--no-allow-dirty"])
            .run()
            .expect(1)
        )
        assert "is unclean" in result.stderr_s

    class TestLocked:
        @pytest.fixture(autouse=True)
        def init(self, nix: Nix, files: Path, hg: Hg):
            nix.nix(["flake", "lock", f"hg+file://{files}/flake-hg2"]).run().ok()
            hg(
                "commit",
                "--config",
                "ui.username=foobar@example.org",
                files / "flake-hg2",
                "-m",
                "Add lock file",
            )

        def test_metadata_clean(self, nix: Nix, files: Path):
            metadata = (
                nix.nix(["flake", "metadata", "--json", f"hg+file://{files}/flake-hg2"])
                .run()
                .json()
            )
            assert "revision" in metadata

        def test_metadata_refresh(self, nix: Nix, files: Path):
            metadata = (
                nix.nix(
                    ["flake", "metadata", "--json", f"hg+file://{files}/flake-hg2", "--refresh"]
                )
                .run()
                .json()
            )
            assert "revision" in metadata
            assert metadata["revCount"] == 2

        @pytest.mark.parametrize("arg", ["--no-registries", "--no-use-registries"])
        def test_build(self, nix: Nix, files: Path, arg: str):
            nix.nix(["build", f"hg+file://{files}/flake-hg2", arg, "--no-allow-dirty"]).run().ok()
            assert (nix.env.dirs.home / "result/hello").exists()



# copy pasta from libutil-support's terminal-code-eater.cc

import enum
import dataclasses


class State(enum.Enum):
    ExpectESC = 1
    ExpectESCSeq = 2
    InCSIParams = 3
    InCSIIntermediates = 4
    InOSCParams = 5
    InOSCST = 6


@dataclasses.dataclass
class TerminalCodeEater:
    state: State = State.ExpectESC

    def feed(self, data: bytes) -> bytes:
        def is_param_char(char: int) -> bool:
            return 48 <= char <= 63

        def is_intermediate_char(char: int) -> bool:
            return 32 <= char <= 47

        def is_final_char(char: int) -> bool:
            return 64 <= char <= 126

        ret = bytearray()
        for c in data:
            match self.state:
                case State.ExpectESC:
                    match c:
                        case 0x1B:  # \e
                            self._transition(State.ExpectESCSeq)
                            continue
                        case 0xD:  # \r
                            continue
                    ret.append(c)
                case State.ExpectESCSeq:
                    match c:
                        case 0x5B:
                            # corresponds to CSI ('[')
                            self._transition(State.InCSIParams)
                            continue
                        case 0x5D:
                            # corresponds to OSC (']')
                            self._transition(State.InOSCParams)
                            continue
                        # FIXME(jade): whatever this was, we do not know how to
                        #  delimit it, so we just eat the next character and
                        #  keep going. Should we actually eat it?
                        case _:
                            self._transition(State.ExpectESC)
                            continue
                # https://en.wikipedia.org/wiki/ANSI_escape_code#CSI_(Control_Sequence_Introducer)_sequences
                # A CSI sequence is: CSI [\x30-\x3f]* [\x20-\x2f]* [\x40-\x7e]
                #                        ^ params     ^ intermediates ^ final byte
                case State.InCSIParams:
                    if is_final_char(c):
                        self._transition(State.ExpectESC)
                        continue
                    if is_intermediate_char(c):
                        self._transition(State.InCSIIntermediates)
                        continue
                    if is_param_char(c):
                        continue
                    msg = f"Corrupt escape sequence, at {c:x}"
                    raise ValueError(msg)
                case State.InCSIIntermediates:
                    if is_final_char(c):
                        self._transition(State.ExpectESC)
                        continue
                    if is_intermediate_char(c):
                        continue
                    msg = f"Corrupt escape sequence in intermediates, at {c:x}"
                    raise ValueError(msg)
                # An OSC is OSC [\x20-\x7e]* ST per ECMA-48
                # where OSC is \x1b ] and ST is \x1b \.
                case State.InOSCParams:
                    # first part of ST
                    if c == 0x1B:
                        self._transition(State.InOSCST)
                        continue
                    # OSC sequences can be ended by BEL on old xterms
                    if c == 0x07:
                        self._transition(State.ExpectESC)
                        continue
                    if c < 0x20 or c == 0x7F:
                        msg = f"Corrupt OSC sequence, at {c:x}"
                        raise ValueError(msg)
                    # either way, eat it
                    continue
                case State.InOSCST:
                    # ST ends by \
                    if c == 0x5C:  # \
                        self._transition(State.ExpectESC)
                    elif c < 0x20 or c > 0x7E:
                        msg = f"Corrupt OSC sequence in ST, at {c:x}"
                        raise ValueError(msg)
                    else:
                        self._transition(State.InOSCParams)
                    continue

        return bytes(ret)

    def _transition(self, new_state: State):
        self.state = new_state


def eat_terminal_codes(s: bytes) -> bytes:
    return TerminalCodeEater().feed(s)



from testlib.fixtures.nix import Nix
from testlib.fixtures.git import Git
from testlib.fixtures.file_helper import with_files, File
from testlib.utils import get_global_asset_pack
from pathlib import Path


@with_files(
    {
        "flake1": get_global_asset_pack(".git")
        | {
            "flake.nix": File("""{
                outputs = { self }: { x = import ./x.nix; };
            }"""),
            "x.nix": File("123"),
        },
        "flake2": get_global_asset_pack(".git")
        | {
            "flake.nix": File("""{
                outputs = { self, flake1 }: { x = flake1.x; };
            }""")
        },
    }
)
def test_unlocked_override(nix: Nix, files: Path, git: Git):
    nix.settings.add_xp_feature("nix-command", "flakes")

    git(files / "flake1", "add", "flake.nix", "x.nix")
    git(files / "flake1", "commit", "-m", "Initial")

    git(files / "flake2", "add", "flake.nix")

    assert (
        nix.nix(
            ["eval", "--json", f"{files}/flake2#x", "--override-input", "flake1", files / "flake1"]
        )
        .run()
        .ok()
        .stdout_s
        == "123\n"
    )

    (files / "flake1/x.nix").write_text("456")

    assert (
        nix.nix(
            ["eval", "--json", f"{files}/flake2#x", "--override-input", "flake1", files / "flake1"]
        )
        .run()
        .ok()
        .stdout_s
        == "456\n"
    )



from pathlib import Path
import sqlite3

import aiohttp.web as web
import pytest

from testlib.fixtures.file_helper import File, with_files
from testlib.fixtures.http_server import http_server
from testlib.fixtures.nix import Nix, with_diverted_store


class HTTPStore:
    def __init__(self):
        self.uploaded_nars = {}
        self.known_nar_hashes = set()

    async def upload_narinfo(self, req: web.Request) -> web.Response:
        self.uploaded_nars[req.match_info["hash"]] = self._parse_narinfo(await req.text())
        return web.Response(text="")

    async def upload_nar(self, req: web.Request) -> web.Response:
        self.known_nar_hashes.add(req.match_info["narhash"])
        return web.Response(text="")

    async def nix_cache_info(self, _: web.Request) -> web.Response:
        return web.Response(text="StoreDir: /nix/store")

    async def get_narinfo(self, req: web.Request) -> web.Response:
        narinfo_hash = req.match_info["hash"]
        if narinfo_hash not in self.uploaded_nars:
            return web.Response(text="", status=404)
        return web.Response(
            text="\n".join(f"{k}: {v}" for k, v in self.uploaded_nars[narinfo_hash].items()) + "\n"
        )

    async def nar_head(self, req: web.Request) -> web.Response:
        return web.Response(
            text="", status=200 if req.match_info["narhash"] in self.known_nar_hashes else 404
        )

    def _parse_narinfo(self, text: str) -> dict[str, str]:
        narinfo = {}
        for line in text.splitlines():
            key, value = line.split(": ", 1)
            narinfo[key] = value
        _, hashpart = narinfo["FileHash"].split(":", 1)
        assert hashpart in self.known_nar_hashes
        return narinfo


class FakeNARBridge(HTTPStore):
    """
    HTTP Store that mutates the URL field of the narinfo, just like
    nar-bridge from snix.dev.

    Testcase to ensure that the correct URL to the nar (i.e. nar/snix-castore/...)
    ends up in the disk-cache.
    """

    async def upload_narinfo(self, req: web.Request) -> web.Response:
        narinfo = self._parse_narinfo(await req.text())
        narinfo["URL"] = (
            f"nar/snix-castore/00000000000000000000000000000000000000000000000000000?narsize=f{narinfo['FileSize']}"
        )

        self.uploaded_nars[req.match_info["hash"]] = narinfo
        return web.Response(text="")


@pytest.fixture(params=[HTTPStore, FakeNARBridge])
def store(request: pytest.FixtureRequest) -> HTTPStore:
    store_class = request.param
    return store_class()


def start_server(store: HTTPStore) -> web.Application:
    app = web.Application()
    app.add_routes(
        [
            web.put("/{hash}.narinfo", store.upload_narinfo),
            web.get("/{hash}.narinfo", store.get_narinfo),
            web.put("/nar/{narhash}.nar", store.upload_nar),
            web.get("/nix-cache-info", store.nix_cache_info),
            web.head("/nar/{narhash}.nar", store.nar_head),
        ]
    )

    return app


def nars_from_narinfo_cache(db_path: Path) -> list[dict[str, str | bool]]:
    assert db_path.exists()
    db = sqlite3.connect(db_path)
    rows = db.execute(
        "SELECT n.present, n.hashPart, n.namePart, n.url FROM NARs n INNER JOIN BinaryCaches b ON n.cache = b.id WHERE b.url LIKE '%localhost%'"
    )
    return [
        {"present": bool(present), "hashPart": hashPart, "namePart": namePart, "url": url}
        for present, hashPart, namePart, url in rows
    ]


@with_files({"test-file": File("hello world")})
@with_diverted_store
def test_http_simple(nix: Nix, store: HTTPStore, files: Path):
    test_file = files / "test-file"
    result = nix.nix(cmd=["store", "add-file", test_file], flake=True).run()
    result.ok()
    store_path = result.stdout_plain
    hash_part, _ = Path(store_path).stem.split("-", 1)

    nar_info_cache = nix.env.dirs.xdg_cache_home / "nix" / "binary-cache-v6.sqlite"

    app = start_server(store)
    with http_server(app) as httpd:
        url = f"http://localhost:{httpd.port}?compression=none&store=/nix/store"
        nix.nix(cmd=["store", "ping", "--store", url], flake=True).run().ok()

        # Narinfo shouldn't exist yet
        nix.nix(cmd=["path-info", "--store", url, store_path], flake=True).run().expect(1)
        cache_entries = nars_from_narinfo_cache(nar_info_cache)

        assert len(cache_entries) == 1
        assert not cache_entries[0]["present"]
        assert cache_entries[0]["hashPart"] == hash_part

        # Successful upload
        nix.nix(
            cmd=["copy", "--from", nix.settings.store, "--to", url, store_path], flake=True
        ).run().ok()
        assert hash_part in store.uploaded_nars

        # Make sure the negative entry got removed
        cache_entries = nars_from_narinfo_cache(nar_info_cache)
        assert len(cache_entries) == 0

        # Ensure that the narinfo can be found now.
        nix.nix(cmd=["path-info", "--store", url, store_path], flake=True).run().ok()

        # Ensure local narinfo cache is up-to-date.
        nar_entries = nars_from_narinfo_cache(nar_info_cache)
        assert len(nar_entries) == 1

        assert nar_entries[0]["present"]
        assert nar_entries[0]["hashPart"] == hash_part
        assert nar_entries[0]["namePart"] == "test-file"
        assert nar_entries[0]["url"] == store.uploaded_nars[hash_part]["URL"]



from collections.abc import Callable
from pathlib import Path

from testlib.fixtures.file_helper import with_files, File, AssetSymlink, Symlink
from testlib.fixtures.nix import Nix
from testlib.fixtures.snapshot import Snapshot


@with_files(
    {
        "in.nix": File("import symlink-resolution/foo/overlays/overlay.nix"),
        "out.exp": AssetSymlink("assets/symlink-resolution.out.exp"),
        "symlink-resolution": {
            "foo": {"lib": {"default.nix": File('"test"')}, "overlays": Symlink("../overlays")},
            "overlays": {"overlay.nix": File("import ../lib")},
        },
    }
)
def test_symlink_resolution(nix: Nix, files: Path, snapshot: Callable[[str], Snapshot]):
    res = nix.nix_instantiate(["--eval", "--strict", files / "in.nix"]).run().ok()
    assert snapshot("out.exp") == res.stdout_plain
    assert not res.stderr_plain



import json
import re

import pytest

from testlib.fixtures.file_helper import CopyFile, with_files
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset

_files = {
    "search.nix": CopyFile("assets/test_search/search.nix"),
    "config.nix": get_global_asset("config.nix"),
}
_exclude_args = ["search", "-f", "search.nix"]
_search_args = [*_exclude_args, ""]


@pytest.fixture(autouse=True)
def commands(nix: Nix):
    nix.settings.add_xp_feature("nix-command")


@with_files(_files)
def test_pkg_name(nix: Nix):
    res = nix.nix([*_search_args, "hello"]).run().ok()
    assert len(res.stdout_plain.splitlines()) > 1


@with_files(_files)
def test_description(nix: Nix):
    res = nix.nix([*_search_args, "broken"]).run().ok()
    assert len(res.stdout_plain.splitlines()) > 1


@with_files(_files)
def test_no_exist(nix: Nix):
    res = nix.nix([*_search_args, "nosuchpackageexits"]).run().expect(1)
    assert "error: no results for the given search" in res.stderr_plain


@with_files(_files)
def test_multiple_args_exists(nix: Nix):
    res = nix.nix([*_search_args, "hello", "empty"]).run().ok()
    assert len(res.stdout_plain.splitlines()) == 2


@with_files(_files)
def test_multiple_args_no_exist(nix: Nix):
    res = nix.nix([*_search_args, "hello", "broken"]).run().expect(1)
    assert "error: no results for the given search" in res.stderr_plain


@with_files(_files)
def test_no_args(nix: Nix):
    res = nix.nix([*_search_args]).run().expect(1)
    assert "error: Must provide at least one regex!" in res.stderr_plain


@with_files(_files)
def test_match_all(nix: Nix):
    res = nix.nix([*_search_args, "^"]).run().ok()
    out = res.stdout_plain
    for pkg in ["foo", "bar", "hello"]:
        assert pkg in out


# FIXME(Jade): possibly not test this with colour in the future
@with_files(_files)
def test_overlap_1(nix: Nix):
    nix.env.set_env("FORCE_COLOR", "1")
    res = nix.nix([*_search_args, "oo", "foo", "oo"]).run().ok()
    out = res.stdout
    assert out == b"* \x1b[0;1m\x1b[32;1mfoo\x1b[0;1m\x1b[0m (5)\n"


@with_files(_files)
def test_overlap_2(nix: Nix):
    nix.env.set_env("FORCE_COLOR", "1")
    res = nix.nix([*_search_args, "broken b", "en bar"]).run().ok()
    out = res.stdout
    assert out == b"* \x1b[0;1mbar\x1b[0m (3)\n  \x1b[32;1mbroken bar\x1b[0m\n"


@with_files(_files)
def test_multiple_finds_o(nix: Nix):
    """Searching for 'o' should yield the 'o' in 'broken bar', the 'oo' in foo and 'o' in hello"""
    nix.env.set_env("FORCE_COLOR", "1")
    res = nix.nix([*_search_args, "o"]).run().ok()
    out = res.stdout_s
    assert len(re.findall(r"\x1b\[32;1mo{1,2}\x1b\[(0|0;1)m", out)) == 3


@with_files(_files)
def test_multiple_finds_b(nix: Nix):
    """Searching for 'b' should yield the 'b' in bar and the two 'b's in 'broken bar'"""
    nix.env.set_env("FORCE_COLOR", "1")
    res = nix.nix([*_search_args, "b"]).run().ok()
    out = res.stdout_s
    assert len(re.findall(r"\x1b\[32;1mb\x1b\[(0|0;1)m", out)) == 3


@with_files(_files)
def test_exclude_hello(nix: Nix):
    res = nix.nix([*_exclude_args, "", "^", "-e", "hello"]).run().ok()
    assert res.stdout_plain.count("hello") == 0


@with_files(_files)
def test_exclude_foo_regex(nix: Nix):
    res = nix.nix([*_exclude_args, "foo", "^", "--exclude", "foo|bar"]).run().expect(1)
    assert "error: no results for the given search" in res.stderr_plain


@with_files(_files)
def test_exclude_foo_multiple_args(nix: Nix):
    res = nix.nix([*_exclude_args, "foo", "^", "-e", "foo", "--exclude", "bar"]).run().expect(1)
    assert "error: no results for the given search" in res.stderr_plain


@with_files(_files)
def test_exclude_bar(nix: Nix):
    res = nix.nix([*_exclude_args, "", "^", "-e", "bar", "--json"]).run().ok()
    assert json.loads(res.stdout_plain).keys() == {"foo", "hello"}



from pathlib import Path

from testlib.fixtures.nix import Nix
from testlib.fixtures.file_helper import CopyFile, Symlink, with_files, File
from testlib.utils import get_global_asset_pack
from testlib.fixtures.git import Git


@with_files(
    {"flake1": {"flake.nix": CopyFile("assets/flake1.nix")}, "flake1_sym": Symlink("flake1")}
)
def test_symlink_points_to_flake(nix: Nix, files: Path) -> None:
    result = nix.nix(["eval", f"{files}/flake1_sym#x"], flake=True).run().ok()
    assert result.stdout_plain == "2"


@with_files(
    {
        "subdir": {"flake1": {"flake.nix": CopyFile("assets/flake1.nix")}},
        "subdir_sym": Symlink("subdir"),
    }
)
def test_symlink_points_to_flake_in_subdir(nix: Nix, files: Path) -> None:
    result = nix.nix(["eval", f"{files}/subdir_sym/flake1#x"], flake=True).run().ok()
    assert result.stdout_plain == "2"


@with_files(
    {
        "flake1": {"subdir": {"flake.nix": CopyFile("assets/flake1.nix")}},
        "flake1_sym": Symlink("flake1/subdir"),
    }
)
def test_symlink_points_to_dir_in_repo(nix: Nix, files: Path) -> None:
    result = nix.nix(["eval", f"{files}/flake1_sym#x"], flake=True).run().ok()
    assert result.stdout_plain == "2"


@with_files(
    {
        "repo1": get_global_asset_pack(".git")
        | {"file": File("Hello"), "subdir": {"flake.nix": CopyFile("assets/flake2.nix")}},
        "repo2": get_global_asset_pack(".git")
        | {"file": File("World"), "flake1_sym": Symlink("../repo1/subdir")},
    }
)
def test_symlink_from_repo_to_another(nix: Nix, files: Path, git: Git) -> None:
    git(files / "repo1", "add", "subdir/flake.nix", "file")
    result = nix.nix(["eval", f"{files}/repo1/subdir#x"], flake=True).run().ok()
    assert result.stdout_plain == '"Hello"'

    git(files / "repo2", "add", "flake1_sym", "file")
    result = nix.nix(["eval", f"{files}/repo2/flake1_sym#x"], flake=True).run().ok()
    assert result.stdout_plain == '"Hello"'


@with_files(
    {
        "repo1": get_global_asset_pack(".git")
        | {"flake.nix": CopyFile("assets/flake_a.nix"), "subdir": {}},
        "repo2": get_global_asset_pack(".git")
        | {"flake.nix": CopyFile("assets/flake_b.nix"), "subdir_sym": Symlink("../repo1/subdir")},
    }
)
def test_symlink_to_subdir_without_flake(nix: Nix, files: Path, git: Git) -> None:
    git(files / "repo1", "add", "flake.nix", "subdir")
    git(files / "repo2", "add", "flake.nix", "subdir_sym")

    result = nix.nix(["eval", f"{files}/repo2/subdir_sym#x"], flake=True).run().ok()
    assert result.stdout_plain == '"b"'



from pathlib import Path
from textwrap import dedent

import pytest
from testlib.fixtures.file_helper import (
    CopyFile,
    CopyTemplate,
    CopyTree,
    File,
    Symlink,
    AssetSymlink,
    merge_file_declaration,
    with_files,
)


@with_files({"test_file_1": File("this is some test content")})
def test_file_creates(files: Path):
    file_name = "test_file_1"
    file_content = "this is some test content"

    target_path = files / file_name
    assert target_path.exists()
    assert target_path.read_text() == file_content


@with_files({"test_file_2": File("asdf")})
def test_no_file_leakage(files: Path):
    file_name = "test_file_2"
    file_content = "asdf"

    some_other_file_name = "test_file_1"

    assert (files / file_name).exists()
    assert (files / file_name).read_text() == file_content
    # Check for no leakage from other tests
    assert not (files / some_other_file_name).exists()


@with_files({"copy_file_test.txt": CopyFile("assets/test_file_helper/copy_file_test.txt")})
def test_copy_file_no_name(files: Path):
    target_path = files / "copy_file_test.txt"

    assert target_path.exists()

    assert target_path.read_text() == "some amazing content\n"


@with_files({"new_name.txt": CopyFile("assets/test_file_helper/copy_file_test.txt")})
def test_copy_file_with_name(files: Path):
    assert (files / "new_name.txt").exists()
    assert not (files / "copy_file_test.txt").exists()


@with_files(
    {
        "target.txt": CopyTemplate(
            "assets/test_file_helper/copy_file_template.template",
            {"some_key": "abc", "other_key": 123, "key_in_braces": "in_braces"},
        )
    }
)
def test_template(files: Path):
    assert (files / "target.txt").exists()

    expected_content = dedent("""
        some_key is set to abc
        more of abc
        other_key is 123
        @not_a_key@
        @in_braces@
    """)

    actual_content = (files / "target.txt").read_text()
    assert actual_content == expected_content


@with_files(
    {
        "target.txt": CopyTemplate(
            "assets/test_file_helper/copy_file_template.template",
            {"some_key": "abc", "key_in_braces": "in_braces"},
        )
    }
)
@pytest.mark.xfail(raises=KeyError)
def test_template_missing_key(files: Path):
    # Empty, because the initialization of the CopyTemplate fails. Caught and tested for by the xfail mark
    ...


@with_files(
    {
        "target.txt": CopyTemplate(
            "assets/test_file_helper/copy_file_template.template",
            {
                "some_key": "Eragon",
                "key_in_braces": "in_braces",
                "other_key": "Arthur Leywin",
                "the beginning": "after the end",
            },
        )
    }
)
@pytest.mark.xfail(raises=ExceptionGroup)
def test_template_too_many_keys(files: Path):
    # Empty, because the initialization of the CopyTemplate fails. Caught and tested for by the xfail mark
    ...


@with_files({"some_folder": CopyTree("assets/test_file_helper/test_folder")})
def test_copy_tree(files: Path):
    files = files / "some_folder"
    assert (files / "a.txt").exists()
    assert (files / "b.txt").exists()
    assert (files / "sub_test").exists()
    assert (files / "sub_test" / "c.txt").exists()


@with_files(
    {
        "a.txt": File("Hello World"),
        "assembler.txt": File("reinforced iron plates\n"),
        "folder_1": {
            "empty.nix": File(""),
            "folder_2": {"not_empty.py": File("print('Arrruuuuuuuuuuuu')")},
        },
    }
)
def test_files_sub_dirs(files: Path):
    f = files / "a.txt"
    assert f.exists()
    assert f.read_text() == "Hello World"

    f = files / "assembler.txt"
    assert f.exists()

    fldr = files / "folder_1"
    assert fldr.exists()
    f = fldr / "empty.nix"
    assert f.exists()

    fldr /= "folder_2"
    assert fldr.exists()

    f = fldr / "not_empty.py"
    assert f.exists()


@with_files({"not_empty": {"a.txt": File("Hello World")}, "empty": {}})
def test_creates_empty_directory(files: Path):
    assert (files / "not_empty").exists()
    assert (files / "empty").exists()
    assert (files / "empty").is_dir()


@with_files({"a.sh": File("echo test", mode=0o477)})
def test_mode_setting(files: Path):
    file = files / "a.sh"
    assert file.exists()
    assert file.stat().st_mode & 0o477 == 0o477


@with_files({"a": AssetSymlink("assets/test_file_helper/copy_file_test.txt")})
def test_asset_symlink(files: Path):
    file = files / "a"
    assert file.exists(follow_symlinks=False)

    assert file.is_symlink()

    assert file.readlink().exists()

    assert file.read_text() == "some amazing content\n"


@with_files({"a": AssetSymlink("assets/test_file_helper/test_folder")})
def test_dir_symlink(files: Path):
    folder = files / "a"
    assert folder.exists(follow_symlinks=False)

    assert folder.is_symlink()

    assert folder.readlink().exists()

    assert folder.readlink().is_dir()

    assert (folder / "a.txt").exists()


@with_files({"a": AssetSymlink("assets/test_file_helper/this_does_not_exist")})
def test_invalid_asset_symlink(files: Path):
    link = files / "a"
    assert link.exists(follow_symlinks=False)
    assert link.is_symlink()
    assert not link.readlink().exists()


@pytest.mark.xfail(raises=ValueError)
@with_files({"a": AssetSymlink("/absolute/assets/test_file_helper/this_does_not_exist")})
def test_absolute_asset_symlink(files: Path):
    pass


@with_files({"tg": File("Hello World"), "folder": {"link": Symlink("../tg")}})
def test_file_symlink(files: Path):
    assert (files / "tg").exists()
    link = files / "folder" / "link"
    assert link.exists(follow_symlinks=False)
    assert link.is_symlink()
    # check that this is actually relative and not an absolute path
    assert str(link.readlink()) == "../tg"


@with_files({"test-file.txt": File("a")}, {"test-file.txt": File("b")})
def test_multiple_file_sets(files: Path):
    file = files / "test-file.txt"
    assert file.exists()
    assert file.read_text() == "a" or file.read_text() == "b"


def test_merge_fd_merges_correctly():
    fa = File("a")
    fc = File("c")
    fd = File("d")
    ff = File("f")
    fd1 = {"a": fa, "b": {"c": fc}}

    fd2 = {"b": {"d": fd}, "e": {"f": ff}}

    expected = {"a": fa, "b": {"c": fc, "d": fd}, "e": {"f": ff}}

    result = merge_file_declaration(fd1, fd2)
    assert result == expected


def test_merge_fd_throws_on_conflict():
    fd1 = {"a": {"b": File("b")}}
    fd2 = {"a": File("a")}

    with pytest.raises(
        ValueError,
        match=r"\('Cannot merge files; got two different values for the same path', 'a'\)",
    ):
        merge_file_declaration(fd1, fd2)


def test_merge_fd_throws_on_conflict_with_full_path():
    fd1 = {"a": {"b": {"c": {"d": File("d")}}}}
    fd2 = {"a": {"b": {"c": {"d": File("different file")}}}}

    with pytest.raises(
        ValueError,
        match=r"\('Cannot merge files; got two different values for the same path', 'a/b/c/d'\)",
    ):
        merge_file_declaration(fd1, fd2)



#!/usr/bin/env python3

"""
This script is a helper for this project's Meson buildsystem, to replace its
usage of `nix eval --write-to`. Writing a JSON object as a nested directory
tree is more generic, easier to maintain, and far, far less cursed. Nix
has 'good' support for JSON output. Let's just use it.
"""

import argparse
from pathlib import Path
import json
import sys

name = 'json-to-tree.py'

def log(*args, **kwargs):
    kwargs['file'] = sys.stderr
    return print(f'{name}:', *args, **kwargs)

def write_dict_to_directory(current_directory: Path, data: dict, files_written=0):
    current_directory.mkdir(parents=True, exist_ok=True)
    for key, value in data.items():
        nested_path = current_directory / key
        match value:
            case dict(nested_data):
                files_written += write_dict_to_directory(nested_path, nested_data)

            case str(content):
                nested_path.write_text(content)
                files_written += 1

            case rest:
                assert False, \
                    f'should have been called on a dict or string, not {type(rest)=}\n\t{rest=}'

    return files_written

def main():
    parser = argparse.ArgumentParser(name)
    parser.add_argument('-i', '--input', type=argparse.FileType('r'), default='-',
        help='The JSON input to operate on and output as a directory tree',
    )
    parser.add_argument('-o', '--output', type=Path, required=True,
        help='The place to put the directory tree',
    )
    args = parser.parse_args()

    json_string = args.input.read()

    try:
        data = json.loads(json_string)
    except json.JSONDecodeError:
        log(f'could not decode JSON from input: {json_string}')
        raise


    files_written = write_dict_to_directory(args.output, data)
    log(f'wrote {files_written} files')

sys.exit(main())



import json

version_json = json.load(open('version.json'))
VERSION = version_json['version']
MAJOR = '.'.join(VERSION.split('.')[:2])
RELEASE_NAME = version_json['release_name']
OFFICIAL_RELEASE = version_json['official_release']



import dataclasses
from enum import Enum
from textwrap import dedent
from typing import ClassVar, NamedTuple

from common import cxx_literal, generate_file, load_data, get_argument_parser


class FeatureTypeNames(NamedTuple):
    code_tag: str
    doc_tag: str


class TimelineEvent(NamedTuple):
    date: str
    release: str
    message: str
    cls: list[int]


class FeatureType(FeatureTypeNames, Enum):
    experimental = FeatureTypeNames("Xp", "xp")
    deprecated = FeatureTypeNames("Dep", "dp")


@dataclasses.dataclass
class ExtraFeature:
    name: str
    internal_name: str
    documentation: str
    timeline: list[TimelineEvent] = dataclasses.field(default_factory=list)

    type: ClassVar[FeatureType]

    @property
    def code(self) -> str:
        return dedent(f"""
            {{
                .tag = {ExtraFeature.type.code_tag}::{self.internal_name},
                .name = {cxx_literal(self.name)},
                .description = {cxx_literal(self.documentation)},
            }},
        """)

    @property
    def docs(self) -> str:
        timeline = (
            f"""

            ### Timeline

            {
                "\n            ".join(
                    [
                        f"- {event.date}, {event.release}: {event.message} [{", ".join([f'[CL {cl}](https://git.lix.systems/c/lix/+/{cl})' for cl in event.cls])}]"
                        for event in self.timeline
                    ]
                )
            }
        """
            if self.timeline
            else ""
        )
        return dedent(f"""
            ## [`{self.name}`]{{#{ExtraFeature.type.doc_tag}-feature-{self.name}}}

            {self.documentation.replace("\n", f"\n{'    ' * 3}")}

            {timeline}

        """)

    @property
    def short_docs(self) -> str:
        return f"  - [`{self.name}`](@docroot@/contributing/{ExtraFeature.type.name}-features.md#{ExtraFeature.type.doc_tag}-feature-{self.name})\n"


def main():
    ap = get_argument_parser()
    ap.add_argument("--deprecated", action="store_true", help="Generate deprecated features")
    ap.add_argument("--impl-header", help="Path of the implementation header to generate")
    ap.add_argument("--shortlist", help="Path of the shortlist file to generate")
    args = ap.parse_args()

    ExtraFeature.type = FeatureType.deprecated if args.deprecated else FeatureType.experimental

    def load(**kwargs) -> ExtraFeature:
        kwargs["timeline"] = [TimelineEvent(**args) for args in kwargs.get("timeline", [])]
        return ExtraFeature(**kwargs)

    features = load_data(args.defs, load)

    generate_file(
        args.header,
        features,
        lambda feature: feature.name,
        lambda feature: f"    {feature.internal_name},\n",
    )
    generate_file(args.impl_header, features, lambda feature: feature.name, lambda f: f.code)
    generate_file(args.docs, features, lambda feature: feature.name, lambda f: f.docs)
    generate_file(args.shortlist, features, lambda feature: feature.name, lambda f: f.short_docs)


if __name__ == "__main__":
    main()



import logging
from logging import Logger

import pytest
from _pytest.fixtures import FixtureRequest


@pytest.fixture
def logger(request: FixtureRequest) -> Logger:
    """
    Returns a logger object to use in a test function.
    :param request: provide by pytest, information about where this fixture was used
    :return: a logger object to use
    """
    return logging.getLogger(request.function.__name__)



from xonsh.main import setup
import signal

# This is a workaround for https://github.com/xonsh/xonsh/issues/5244
# Matching https://github.com/xonsh/xonsh/blob/71d4920ac5c16f2ef37da88b3f563c25c585fb82/xonsh/main.py#L519-L525
def func_sig_ttin_ttou(n, f):
    pass

signal.signal(signal.SIGTTIN, func_sig_ttin_ttou)
signal.signal(signal.SIGTTOU, func_sig_ttin_ttou)

setup()
del setup

import logging
import sys

import xonsh.shells.base_shell

from . import environment
from . import create_release
from . import keys
from . import version
from . import cli
from . import docker
from . import docker_assemble
from . import release_notes
from . import gitutils


def setup_logging():
    """
    Sets up logging to work properly. The following are intended to work:
    - ipython/xonsh configuration files adding log handlers out of band
    - Reloading the module in xonsh/ipython not causing Bonus Loggers (which is
      why we check if there is already a handler. This also helps the previous
      case)
    - Importing the releng module from xonsh and poking at it interactively
    """

    LEVELS = {
        # Root logger must be DEBUG so that anything else can be DEBUG
        None: logging.DEBUG,
        # Everything in releng
        __name__: logging.DEBUG,
        # Log spam caused by prompt_toolkit
        'asyncio': logging.INFO,
    }

    for name, level in LEVELS.items():
        logger = logging.getLogger(name)
        logger.setLevel(level)

    root_logger = logging.getLogger()

    fmt = logging.Formatter('{asctime} {levelname} {name}: {message}',
                            datefmt='%b %d %H:%M:%S',
                            style='{')

    if not any(
            isinstance(h, logging.StreamHandler) for h in root_logger.handlers):
        stderr = sys.stderr
        # XXX: Horrible hack required by the virtual stderr xonsh uses for each entered
        # command getting closed after the command is run: we need to pull out
        # the real stderr because this survives across multiple command runs.
        #
        # This only applies when running xonsh in interactive mode and importing releng.
        if isinstance(sys.stderr, xonsh.shells.base_shell._TeeStd):
            stderr = stderr.std  # type: ignore

        hand = logging.StreamHandler(stream=stderr)
        hand.set_name('releng root handler')
        hand.setFormatter(fmt)
        root_logger.addHandler(hand)


setup_logging()


def reload():
    import importlib
    importlib.reload(environment)
    importlib.reload(create_release)
    importlib.reload(keys)
    importlib.reload(version)
    importlib.reload(cli)
    importlib.reload(docker)
    importlib.reload(docker_assemble)
    importlib.reload(gitutils)
    importlib.reload(release_notes)



import string
from string import Template

import pytest


class BalancedTemplater(Template):
    delimiter = "@"
    pattern = r"@((?P<escaped>@)|(?P<named>\w+?)@|\{(?P<braced>\w+?)\}@|(?P<invalid>.*?))"

    def substitute(self, mapping: dict[str, object] | None = None, /, **kwargs) -> str:
        if mapping is None:
            mapping = {}
        tmpl_idents = set(self.get_identifiers())
        subs_idents = mapping.keys() | kwargs.keys()
        errs = []
        for key in subs_idents - tmpl_idents:
            msg = f"Unused named argument `{key}` with value: {kwargs.get(key, mapping.get(key))}"
            errs.append(KeyError(msg))
        if len(errs) > 0:
            msg = "Unused arguments passed to substitute"
            raise ExceptionGroup(msg, errs)
        return super().substitute(mapping, **kwargs)


@pytest.fixture
def balanced_templater() -> type[string.Template]:
    return BalancedTemplater



from xml.etree import ElementTree

from testlib.fixtures.file_helper import with_files
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset_pack


@with_files(get_global_asset_pack("dependencies"))
def test_correct_xml(nix: Nix):
    """Test that nix-env --query --xml produces valid XML."""
    query_output = (
        nix.nix_env(
            ["--query", "--available", "--attr-path", "--xml", "--file", "./dependencies.nix"]
        )
        .run()
        .ok()
        .stdout_plain
    )

    # Should parse without raising.
    parsed = ElementTree.fromstring(query_output)
    assert parsed.tag == "items"



#!/usr/bin/env python3
import subprocess
import argparse


def get_targets_of_rule(build_root: str, rule_name: str) -> list[str]:
    return (
        subprocess.check_output(["ninja", "-C", build_root, "-t", "targets", "rule", rule_name])
        .decode()
        .strip()
        .splitlines()
    )


def ninja_build(build_root: str, targets: list[str]):
    subprocess.check_call(["ninja", "-C", build_root, "--", *targets])


def main():
    ap = argparse.ArgumentParser(description="Builds required targets for clang-tidy")
    ap.add_argument("build_root", help="Ninja build root", type=str)

    args = ap.parse_args()

    targets = (
        [t for t in get_targets_of_rule(args.build_root, "CUSTOM_COMMAND") if t.endswith(".gen.hh")]
        + [
            t
            for t in get_targets_of_rule(args.build_root, "CUSTOM_COMMAND_DEP")
            if t.endswith(".capnp.h")
        ]
        + [
            t
            for t in get_targets_of_rule(args.build_root, "CUSTOM_COMMAND")
            if t.endswith(".gen.inc")
        ]
    )
    ninja_build(args.build_root, targets)


if __name__ == "__main__":
    main()



import pytest
import re

from pathlib import Path

from testlib.fixtures.nix import Nix
from testlib.fixtures.file_helper import with_files
from testlib.utils import get_global_asset
from testlib.fixtures.file_helper import Symlink, File, CopyFile

from typing import Literal

from dataclasses import dataclass


@dataclass
class CertAssessment:
    warnings: list[str]
    outcomes: set[Literal["missing", "present", "present-env-var", "corrupted"]]


certificate_files = {
    "cert": File("CERT_CONTENT"),
    "symlinked-cert": Symlink("cert"),
    "certificate-test.nix": CopyFile("assets/certificate-test.nix"),
    "config.nix": get_global_asset("config.nix"),
}


def assess_cert_presence_in_builds(
    nix: Nix,
    mode: Literal["normal", "fixed-output", "clobbering-impurities"],
    *,
    cert: Path | str | None = None,
    sandboxed: bool = True,
) -> CertAssessment:
    """
    Run a build with some special Nix code that makes use of the certificate authority code.
    We use the stdout to inspect whether our expectations are met.
    We return a list of known metadata about how the build ran.
    """
    expected_ret_code = 1 if mode in ("fixed-output", "clobbering-impurities") else 100

    res = (
        nix.nix_build(
            [
                "certificate-test.nix",
                "--argstr",
                "mode",
                mode,
                "--arg",
                "sandbox",
                "true" if sandboxed else "false",
                "--no-out-link",
            ]
            + (["--option", "ssl-cert-file", str(cert)] if cert is not None else [])
        )
        .run()
        .expect(expected_ret_code)
    )

    contains_clobbering_warning = (
        "warning: 'NIX_SSL_CERT_FILE is an impure environment variable" in res.stderr_plain
    )

    # this should be read as: mode == "clobbering-impurities" implies contains_clobbering_warning
    assert not contains_clobbering_warning or mode == "clobbering-impurities"

    outcomes = set(re.findall(r"CERT_(.*?)_IN_SANDBOX", res.stderr_plain))
    warnings = re.findall(r"warning: (.*)", res.stderr_plain)

    return CertAssessment(outcomes=outcomes, warnings=warnings)


@with_files(certificate_files)
@pytest.mark.parametrize("sandboxed", [False, pytest.param(True, marks=pytest.mark.full_sandbox)])
def test_missing_cert_in_ia_builds(nix: Nix, sandboxed: bool):
    assert assess_cert_presence_in_builds(
        nix, "normal", cert="cert", sandboxed=sandboxed
    ).outcomes == {"missing"}


@with_files(certificate_files)
@pytest.mark.parametrize("cert_path", ["", "/nowhere"])
@pytest.mark.parametrize("sandboxed", [False, pytest.param(True, marks=pytest.mark.full_sandbox)])
def test_missing_cert_in_fod_builds(nix: Nix, sandboxed: bool, cert_path: str):
    assert assess_cert_presence_in_builds(
        nix, "fixed-output", cert=cert_path, sandboxed=sandboxed
    ).outcomes == {"missing"}


@with_files(certificate_files)
@pytest.mark.parametrize("cert_path", ["cert", "symlinked-cert"])
@pytest.mark.parametrize("sandboxed", [False, pytest.param(True, marks=pytest.mark.full_sandbox)])
def test_presence_cert_in_fod_builds(nix: Nix, cert_path: str, sandboxed: bool):
    assert assess_cert_presence_in_builds(
        nix, "fixed-output", cert=cert_path, sandboxed=sandboxed
    ).outcomes == ({"present", "present-env-var"} if sandboxed else {"present-env-var"})

    nix.env["NIX_SSL_CERT_FILE"] = str(cert_path)
    assert assess_cert_presence_in_builds(nix, "fixed-output", sandboxed=sandboxed).outcomes == (
        {"present", "present-env-var"} if sandboxed else {"present-env-var"}
    )


@with_files(certificate_files)
@pytest.mark.parametrize("sandboxed", [False, pytest.param(True, marks=pytest.mark.full_sandbox)])
class TestCertClobberingInFODs:
    def test_no_env(self, nix: Nix, sandboxed: bool):
        # there's no NIX_SSL_CERT_FILE presently, so everything works as usually.
        assert assess_cert_presence_in_builds(
            nix, "clobbering-impurities", cert="cert", sandboxed=sandboxed
        ).outcomes == ({"present", "present-env-var"} if sandboxed else {"present-env-var"})

    def test_with_env_with_flag(self, nix: Nix, sandboxed: bool):
        # NIX_SSL_CERT_FILE is set but we pass --ssl-cert-file which takes precedence.
        # We expect warnings to occur.
        nix.env["NIX_SSL_CERT_FILE"] = "/nowhere"
        assert assess_cert_presence_in_builds(
            nix, "clobbering-impurities", cert="cert", sandboxed=sandboxed
        ).outcomes == ({"present", "present-env-var"} if sandboxed else {"present-env-var"})

    def test_with_env_no_flag(self, nix: Nix, sandboxed: bool):
        # NIX_SSL_CERT_FILE is set but we do not pass any flag.
        # We expect no warning.
        nix.env["NIX_SSL_CERT_FILE"] = "/nowhere"

        assert assess_cert_presence_in_builds(
            nix, "clobbering-impurities", sandboxed=sandboxed
        ).outcomes == {"missing"}

    def test_with_valid_env_no_flag(self, nix: Nix, sandboxed: bool):
        # We change the environment variable to a valid path.
        nix.env["NIX_SSL_CERT_FILE"] = "cert"
        assert assess_cert_presence_in_builds(
            nix, "clobbering-impurities", sandboxed=sandboxed
        ).outcomes == ({"present", "present-env-var"} if sandboxed else {"present-env-var"})

    def test_warning_absence(self, nix: Nix, sandboxed: bool):
        warning_prefix = "'NIX_SSL_CERT_FILE' is an impure environment variable"

        # Under impureEnvVars set, Lix will set `NIX_SSL_CERT_FILE` only if it existed in the environment
        # of the builder.
        # This should cause the warning to be absent.
        nix.env.unset_env("NIX_SSL_CERT_FILE")
        assessment = assess_cert_presence_in_builds(
            nix, "clobbering-impurities", cert="cert", sandboxed=sandboxed
        )

        assert not any(w.startswith(warning_prefix) for w in assessment.warnings)



import shutil

from testlib.fixtures.file_helper import with_files
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset_pack


@with_files(get_global_asset_pack("dependencies"))
def test_dump_db(nix: Nix):
    nix.nix_build(["dependencies.nix", "-o", "result"]).run().ok()

    res = nix.nix_store(["-qR", "result"]).run().ok()
    deps = res.stdout_plain

    res = nix.nix_store(["--dump-db"]).run().ok()
    dump = res.stdout

    shutil.rmtree(nix.env.dirs.nix_state_dir / "db")

    nix.nix_store(["--load-db"]).with_stdin(dump).run().ok()

    res = nix.nix_store(["-qR", "result"]).run().ok()
    deps2 = res.stdout_plain

    assert deps == deps2

    res = nix.nix_store(["--dump-db"]).run().ok()
    dump2 = res.stdout

    assert dump == dump2



from testlib.fixtures.nix import Nix
from testlib.fixtures.file_helper import with_files, File
from testlib.environ import environ
from pathlib import Path
from .common import simple_flake
import pytest

system = environ.get("system")


@pytest.fixture(autouse=True)
def common_init(nix: Nix):
    nix.settings.add_xp_feature("nix-command", "flakes")


@with_files(simple_flake())
class TestFlakeShow:
    def test_default_behavior(self, nix: Nix, files: Path):
        """
        By default: Only show the packages content for the current system and no legacyPackages at all
        """
        result = nix.nix(["flake", "show", "--json", files]).run().ok().json()
        assert result["packages"]["someOtherSystem"]["default"] == {}
        assert result["packages"][system]["default"]["name"] == "simple"
        assert result["legacyPackages"][system] == {}

    def test_eval_system_is_honored(self, nix: Nix, files: Path):
        result = (
            nix.nix(["flake", "show", "--eval-system", "someOtherSystem", "--json", files])
            .run()
            .ok()
            .json()
        )
        assert result["packages"]["someOtherSystem"]["default"]["name"] == "simple"

    def test_all_systems_shows_all(self, nix: Nix, files: Path):
        result = nix.nix(["flake", "show", "--json", "--all-systems", files]).run().ok().json()
        assert result["packages"]["someOtherSystem"]["default"]["name"] == "simple"
        assert result["legacyPackages"][system] == {}

    def test_legacy_flag_shows_legacy_packages(self, nix: Nix, files: Path):
        result = nix.nix(["flake", "show", "--json", "--legacy", files]).run().ok().json()
        assert result["legacyPackages"][system]["hello"]["name"] == "simple"


@with_files(
    {
        "flake.nix": File(f"""{{
          description = "Bla bla";

          outputs = inputs: rec {{
            apps.{system} = {{ }};
            checks.{system} = {{ }};
            devShells.{system} = {{ }};
            legacyPackages.{system} = {{ }};
            packages.{system} = {{ }};
            packages.someOtherSystem = {{ }};

            formatter = {{ }};
            nixosConfigurations = {{ }};
            nixosModules = {{ }};
          }};
        }}""")
    }
)
def test_show_without_contents(nix: Nix, files: Path):
    assert nix.nix(["flake", "show", "--json", "--all-systems", files]).run().ok().json() == {}


@with_files(
    simple_flake()
    | {
        "flake.nix": File(f"""{{
          outputs = inputs: {{
            legacyPackages.{system} = {{
              # nixpkgs.legacyPackages is a particularly prominent source of eval errors
              AAAAAASomeThingsFailToEvaluate = throw "nooo";
              simple = import ./simple.nix;
            }};
          }};
        }}""")
    }
)
def test_show_handles_eval_errors(nix: Nix, files: Path):
    result = (
        nix.nix(["flake", "show", "--json", "--legacy", "--all-systems", files]).run().ok().json()
    )
    assert result["legacyPackages"][system]["AAAAAASomeThingsFailToEvaluate"] == {}
    assert result["legacyPackages"][system]["simple"]["name"] == "simple"



#!/usr/bin/env python3

import json
import os
import subprocess
import pytest
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, List

TEST_ROOT = Path(__file__).parent.resolve()
# subprojects/nix-eval-jobs/tests
# PROJECT_ROOT = TEST_ROOT.parent.parent.parent
# BIN = PROJECT_ROOT.joinpath("outputs", "out", "bin", "nix-eval-jobs")
BIN = "nix-eval-jobs"


def check_gc_root(gcRootDir: str, drvPath: str):
    """
    Make sure the expected GC root exists in the given dir
    """
    link_name = os.path.basename(drvPath)
    symlink_path = os.path.join(gcRootDir, link_name)
    assert os.path.islink(symlink_path) and drvPath == os.readlink(symlink_path)


def evaluate(
    tempdir: TemporaryDirectory,
    expected_statuscode: int = 0,
    extra_args: List[str] = [],
) -> tuple[Dict[str, Dict[str, Any]], str]:
    if nixpkgs_path := os.getenv("NEJ_NIXPKGS_PATH"):
        if "--flake" in extra_args:
            extra_args.extend(["--override-input", "nixpkgs", f"path:{nixpkgs_path}"])
        else:
            extra_args.extend(["--arg", "pkgs", f"import {nixpkgs_path} {{}}"])

    cmd = [
        str(BIN),
        "--gc-roots-dir",
        tempdir,
        "--meta",
        "--extra-experimental-features",
        "flakes",
    ] + extra_args
    res = subprocess.run(
        cmd,
        cwd=TEST_ROOT.joinpath("assets"),
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    assert res.returncode == expected_statuscode
    print(res.stdout)
    print(res.stderr)
    return sorted([json.loads(r) for r in res.stdout.split("\n") if r], key=lambda job: job["attr"]), res.stderr


def common_test(extra_args: List[str]) -> List[Dict[str, Any]]:
    with TemporaryDirectory() as tempdir:
        results, _ = evaluate(tempdir, 0, extra_args)
        assert len(results) == 4

        dotted_job = results[0]
        assert dotted_job["attr"] == '"dotted.attr"'
        assert dotted_job["attrPath"] == ["dotted.attr"]
        check_gc_root(tempdir, dotted_job["drvPath"])

        built_job = results[1]
        assert built_job["attr"] == "builtJob"
        assert built_job["name"] == "job1"
        assert built_job["outputs"]["out"].startswith("/nix/store")
        assert built_job["drvPath"].endswith(".drv")
        assert built_job["meta"]["broken"] is False
        check_gc_root(tempdir, built_job["drvPath"])

        recurse_drv = results[2]
        assert recurse_drv["attr"] == "recurse.drvB"
        assert recurse_drv["name"] == "drvB"
        check_gc_root(tempdir, recurse_drv["drvPath"])

        substituted_job = results[3]
        assert substituted_job["attr"] == "substitutedJob"
        assert substituted_job["name"].startswith("hello-")
        assert substituted_job["meta"]["broken"] is False
        return results


def test_flake() -> None:
    results = common_test(["--flake", ".#hydraJobs"])
    for result in results:
        assert "isCached" not in result


def test_query_cache_status() -> None:
    results = common_test(["--flake", ".#hydraJobs", "--check-cache-status"])
    # FIXME in the nix sandbox we cannot query binary caches
    # this would need some local one
    for result in results:
        assert "isCached" in result


def test_expression() -> None:
    results = common_test(["ci.nix"])
    for result in results:
        assert "isCached" not in result

    with open(TEST_ROOT.joinpath("assets/ci.nix"), "r") as ci_nix:
        common_test(["-E", ci_nix.read()])


def test_eval_error() -> None:
    with TemporaryDirectory() as tempdir:
        results, _ = evaluate(
            tempdir,
            0,
            ["--workers", "1", "--flake", ".#legacyPackages.x86_64-linux.brokenPkgs"],
        )
        assert len(results) == 1

        attr = results[0]
        assert attr["attr"] == "brokenPackage"
        assert "this is an evaluation error" in attr["error"]


@pytest.mark.infiniterecursion
def test_recursion_error() -> None:
    with TemporaryDirectory() as tempdir:
        results, stderr = evaluate(
            tempdir,
            1,
            [
                "--workers",
                "1",
                "--flake",
                ".#legacyPackages.x86_64-linux.infiniteRecursionPkgs",
            ],
        )
        print(stderr)

        assert "packageWithInfiniteRecursion" in stderr
        assert "possible infinite recursion" in stderr


def test_constituents() -> None:
    with TemporaryDirectory() as tempdir:
        results, _ = evaluate(
            tempdir,
            0,
            [
                "--workers",
                "1",
                "--flake",
                ".#legacyPackages.x86_64-linux.constituents.success",
                "--constituents",
            ],
        )
        assert len(results) == 4

        child = results[0]
        assert child["attr"] == "anotherone"
        assert "constituents" not in child
        assert "namedConstituents" not in child

        direct = results[1]
        assert direct["attr"] == "direct_aggregate"
        assert "constituents" in direct
        assert "namedConstituents" not in direct

        indirect = results[2]
        assert indirect["attr"] == "indirect_aggregate"
        assert "constituents" in indirect
        assert "namedConstituents" not in indirect

        mixed = results[3]
        assert mixed["attr"] == "mixed_aggregate"

        def absent_or_empty(f: str, d: dict) -> bool:
            return f not in d or len(d[f]) == 0

        assert absent_or_empty("namedConstituents", direct)
        assert absent_or_empty("namedConstituents", indirect)
        assert absent_or_empty("namedConstituents", mixed)

        assert direct["constituents"][0].endswith("-job1.drv")

        assert indirect["constituents"][0] == child["drvPath"]

        assert mixed["constituents"][0].endswith("-job1.drv")
        assert mixed["constituents"][1] == child["drvPath"]

        assert "error" not in direct
        assert "error" not in indirect
        assert "error" not in mixed

        check_gc_root(tempdir, direct["drvPath"])
        check_gc_root(tempdir, indirect["drvPath"])
        check_gc_root(tempdir, mixed["drvPath"])


def test_constituents_cycle() -> None:
    with TemporaryDirectory() as tempdir:
        results, _ = evaluate(
            tempdir,
            0,
            [
                "--workers",
                "1",
                "--flake",
                ".#legacyPackages.x86_64-linux.constituents.cycle",
                "--constituents",
            ],
        )
        assert len(results) == 2

        assert list(map(lambda x: x["name"], results)) == ["aggregate0", "aggregate1"]
        for i in results:
            assert i["error"] == "Dependency cycle: aggregate0 <-> aggregate1"


def test_constituents_error() -> None:
    with TemporaryDirectory() as tempdir:
        results, _ = evaluate(
            tempdir,
            0,
            [
                "--workers",
                "1",
                "--flake",
                ".#legacyPackages.x86_64-linux.constituents.failures",
                "--constituents",
            ],
        )
        assert len(results) == 2

        aggregate = results[0]
        assert aggregate["attr"] == "aggregate"
        assert "namedConstituents" not in aggregate
        assert "doesntexist: does not exist\n" in aggregate["error"]
        assert "constituents" in aggregate

        child = results[1]
        assert child["attr"] == "doesnteval"
        assert "error" in child


def test_transitivity() -> None:
    with TemporaryDirectory() as tempdir:
        results, _ = evaluate(
            tempdir,
            0,
            [
                "--workers",
                "1",
                "--flake",
                ".#legacyPackages.x86_64-linux.constituents.transitive",
                "--constituents",
            ],
        )
        assert len(results) == 3

        aggregate0 = results[0]
        assert aggregate0["attr"] == "aggregate0"

        aggregate1 = results[1]
        assert aggregate1["attr"] == "aggregate1"

        job = results[2]
        assert job["attr"] == "constituent"
        assert "constituents" not in job

        assert aggregate1["drvPath"] == aggregate0["constituents"][0]


def test_mutually_exclusive_combinations() -> None:
    with TemporaryDirectory() as tempdir:
        for flag in ["constituents", "check-cache-status"]:
            result = subprocess.run(
                [
                    str(BIN),
                    "--gc-roots-dir",
                    tempdir,
                    "--meta",
                    "--extra-experimental-features",
                    "flakes",
                    "--no-instantiate",
                    f"--{flag}",
                    "--workers",
                    "1",
                    "--flake",
                    ".#legacyPackages.x86_64-linux.constituents.success",
                ],
                text=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            assert result.returncode == 1
            assert f"--no-instantiate and --{flag} are mutually exclusive" in result.stderr


def test_no_instantiate_mode() -> None:
    """Test that --no-instantiate flag works correctly"""
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        gcroots = path / "gcroots"
        gcroots.mkdir()
        results, _ = evaluate(
            tempdir,
            0,
            [
                "--gc-roots-dir",
                gcroots,
                "--eval-store",
                path / "root",
                "--meta",
                "--no-instantiate",
                "--flake",
                ".#hydraJobs",
            ]
        )
        assert len(results) == 4

        # Check that all results have the expected structure
        for result in results:
            # In no-instantiate mode, drvPath should still be present (from the attr)
            assert "drvPath" in result
            assert result["drvPath"].endswith(".drv")

            assert not (path / "root" / result["drvPath"][1:]).exists()

            # System should still be present (from querySystem fallback)
            assert "system" in result
            assert result["system"] != ""

            # Name should still be present
            assert "name" in result

            # Outputs should still be present but may be empty
            assert "outputs" in result

            # Cache status should not be present (it's Unknown and not included)
            assert "cacheStatus" not in result
            assert "neededBuilds" not in result
            assert "neededSubstitutes" not in result

            # Input drvs should not be present (requires reading derivation from store)
            assert not result["inputDrvs"]

        # No GC roots should be created in no-instantiate mode
        assert len(list(Path(gcroots).iterdir())) == 0



from pathlib import Path

from testlib.fixtures.file_helper import with_files
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset_pack


@with_files({"simple": get_global_asset_pack("simple-drv")})
def test_path_from_hash_part(nix: Nix):
    nix.settings.add_xp_feature("nix-command")

    path = (
        nix.nix(["build", "--no-link", "--print-out-paths", "-f", "simple/simple.nix"])
        .run()
        .ok()
        .stdout_plain
    )

    hash_part = Path(path).name[0:32]

    path2 = nix.nix(["store", "path-from-hash-part", hash_part]).run().ok().stdout_plain

    assert path == path2
